from typing import Optional
from collections.abc import Iterator
from algoritmia.datastructures.graphs import Digraph, TVertex

def topological_sort(g: Digraph[TVertex]) -> Optional[list[TVertex]]:
    def traverse_from(v) -> Iterator[TVertex]:  # depthfirst postorder
        pending.remove(v)
        for suc in g.succs(v):
            if suc in pending:
                yield from traverse_from(suc)
        yield v

    def next_vertex_no_preds() -> Optional[TVertex]:
        for v in pending:
            if len(g.preds(v)) == 0:
                return v

    pending = set(g.V)
    lv = []
    while len(pending) > 0:
        u = next_vertex_no_preds()
        if u is None:
            return None
        vertices = list(traverse_from(u))
        pending -= set(vertices)
        lv.extend(vertices)
    lv.reverse()
    return lv


if __name__ == '__main__':
    edges = [('C', 'C++'), ('C', 'Java'), ('C', 'Objective-C'), ('C', 'C#'),
             ('C++', 'Java'), ('C++', 'C#'),
             ('Java', 'C#'),
             ('MT', 'Haskell'), ('Haskell', 'C#')]

    my_graph = Digraph(E=edges)

    print(topological_sort(my_graph))



"""
05/10/2021: Version 1.5. Tamaño nodos y margen.
28/09/2021: Version 1.4. Actualizado a easypaint 2.x.
28/09/2021: Version 1.3. Añadidos canvas_width y canvas_height en el constructor
                         (reemplazan a window_size)
15/09/2021: Version 1.2. Permite poner el fondo transparente (en el IU es blanco, pero
                         en el eps se genera transparente).
24/09/2019: Version 1.1. Corregido modo ROW_COL para que funcione como LabyrinthViewer
02/10/2013: Version 1
@author: David Llorens dllorens@uji.es

Dos modos de funcionamiento según parámetro del constructor (vertexmode):

- Modo ROW_COL (8 filas, 14 columnas):
   (0,0)-----(0,14)
     |         |
   (8,0)-----(8,14)

- Modo X_Y (mismos puntos):
   (0,14)--(8,14)
     |       |
     |       |
     |       |
   (0,0)---(8,0)
"""

import tkinter
from math import sin, cos, pi, sqrt
from typing import Optional, Union

from easypaint import EasyPaint

from algoritmia.datastructures.graphs import IGraph, UndirectedGraph, Digraph

Num = Union[int, float]
Vertex = tuple[Num, Num]


# -------------------------------------------------------------------------------------

def dist(p1, p2):
    return sqrt((p1[0] - p2[0]) * (p1[0] - p2[0]) + (p1[1] - p2[1]) * (p1[1] - p2[1]))


def find_intersec_point(cx, cy, radius, p1x, p1y, p2x, p2y):
    dx = p2x - p1x
    dy = p2y - p1y

    A = dx * dx + dy * dy
    B = 2 * (dx * (p1x - cx) + dy * (p1y - cy))
    C = (p1x - cx) * (p1x - cx) + (p1y - cy) * (p1y - cy) - radius * radius

    det = B * B - 4 * A * C
    if A <= 0.0000001 or det < 0:
        raise Exception("No real solutions")
    if det == 0:
        # One solution.
        t = -B / (2 * A)
        return (p1x + t * dx, p1y + t * dy), None
    else:
        # Two solutions.
        t1 = (-B + sqrt(det)) / (2 * A)
        t2 = (-B - sqrt(det)) / (2 * A)
        return (p1x + t1 * dx, p1y + t1 * dy), (p1x + t2 * dx, p1y + t2 * dy)


# -------------------------------------------------------------------------------------


class Graph2dViewer(EasyPaint):
    X_Y = 0
    ROW_COL = 1

    def __init__(self, g: IGraph[Vertex], v_label=None, canvas_width: int = 400, canvas_height: int = 300,
                 margin: int = 15, vertexmode=X_Y, background='white', node_size: Optional[float] = None,
                 title=None, colors: Optional[dict[Vertex, str]] = None):
        self.node_size = node_size
        self.margin = margin
        self.colors = colors if colors is not None else {}
        self.is_directed = isinstance(g, Digraph)
        self.v_label = v_label
        if not isinstance(g, UndirectedGraph) and not isinstance(g, Digraph):
            raise TypeError("The first parameter must be an UndirectedGraph or a Digraph")
        if any([not isinstance(p, tuple) or len(p) != 2 or
                not isinstance(p[0], int) and not isinstance(p[0], float) or
                not isinstance(p[1], int) and not isinstance(p[1], float) for p in g.V]):
            raise TypeError("Vertices must be tuples of two integers or floats")

        if vertexmode == Graph2dViewer.ROW_COL:
            if any([not isinstance(p[0], int) for p in g.V]):
                raise TypeError("In this mode, vertices must be tuples of two integers")
            g = UndirectedGraph(V=[(v[1], -v[0]) for v in g.V], E=[((u[1], -u[0]), (v[1], -v[0])) for (u, v) in g.E])

        self.g = g

        self.max_y = max(p[1] for p in self.g.V)
        self.max_x = max(p[0] for p in self.g.V)
        self.min_y = min(p[1] for p in self.g.V)
        self.min_x = min(p[0] for p in self.g.V)

        super().__init__()

        self.easypaint_configure(size=(canvas_width, canvas_height),
                                 title='2D Graph Viewer' if title is None else title,
                                 background=background)

        w, h = self.size
        w -= self.margin
        h -= self.margin
        sizex = self.max_x - self.min_x + 1
        sizey = self.max_y - self.min_y + 1
        self.cell_size = cell_size = min(w / sizex, h / sizey)

        self.m = ((w - cell_size * sizex) / 2 - self.min_x * cell_size + self.margin / 2,
                  (h - cell_size * sizey) / 2 - self.min_y * cell_size + self.margin / 2)
        if self.node_size is None:
            self.node_size = (w * h / len(self.g.V)) ** 0.5 / 10  # cell_size / 8

    def on_key_press(self, keysym):
        if keysym in ['Return', 'Escape']:
            self.close()

    def draw_arrow(self, u, v, color='black', width=1, tag='arrow'):
        cell_size = self.cell_size
        m = self.m
        if not self.is_directed:
            self.create_line((u[0] + 0.5) * cell_size + m[0], (u[1] + 0.5) * cell_size + m[1],
                             (v[0] + 0.5) * cell_size + m[0], (v[1] + 0.5) * cell_size + m[1])
        else:
            p1, p2 = find_intersec_point(v[0], v[1], (self.node_size + width) / self.cell_size,
                                         u[0], u[1], v[0], v[1])
            _, (x2, y2) = min((dist(u, p1), p1), (dist(u, p2), p2))
            x1, y1 = u
            l2 = self.node_size

            phi = 30 * pi / 180
            l1 = ((x1 - x2) ** 2 + (y1 - y2) ** 2) ** 0.5
            l2 /= self.cell_size

            x3 = x2 + l2 / l1 * ((x1 - x2) * cos(phi) + (y1 - y2) * sin(phi))
            x4 = x2 + l2 / l1 * ((x1 - x2) * cos(phi) - (y1 - y2) * sin(phi))

            y3 = y2 + l2 / l1 * ((y1 - y2) * cos(phi) - (x1 - x2) * sin(phi))
            y4 = y2 + l2 / l1 * ((y1 - y2) * cos(phi) + (x1 - x2) * sin(phi))

            self.create_line((x1 + 0.5) * cell_size + m[0], (y1 + 0.5) * cell_size + m[1],
                             (x2 + 0.5) * cell_size + m[0], (y2 + 0.5) * cell_size + m[1],
                             color, width=width, capstyle=tkinter.ROUND, tag=tag)
            self.create_line((x2 + 0.5) * cell_size + m[0], (y2 + 0.5) * cell_size + m[1],
                             (x3 + 0.5) * cell_size + m[0], (y3 + 0.5) * cell_size + m[1],
                             color, width=width, capstyle=tkinter.ROUND, tag=tag)
            self.create_line((x2 + 0.5) * cell_size + m[0], (y2 + 0.5) * cell_size + m[1],
                             (x4 + 0.5) * cell_size + m[0], (y4 + 0.5) * cell_size + m[1],
                             color, width=width, capstyle=tkinter.ROUND, tag=tag)

    def draw_vertex(self, u: Vertex, color='black', fill='palegreen', width=1, tag='vertex'):
        self.create_filled_circle((u[0] + 0.5) * self.cell_size + self.m[0],
                                  (u[1] + 0.5) * self.cell_size + self.m[1],
                                  self.node_size, color=color, width=width, fill=fill, tag=tag)
        if self.v_label is not None:
            label, fs = self.v_label[u], int(self.node_size)
            self.create_text((u[0] + 0.5) * self.cell_size + self.m[0],
                             (u[1] + 0.5) * self.cell_size + self.m[1],
                             label, fs, color=color, tag=tag)

    def main(self):
        width = 1
        color = 'black'
        for u, v in self.g.E:
            self.draw_arrow(u, v, color=color, width=width)
        for u in self.g.V:
            fill = 'palegreen' if len(self.colors) == 0 else self.colors[u] if u in self.colors else 'white'
            self.draw_vertex(u, fill=fill, width=width)


if __name__ == '__main__':
    c = {'A': (10, 1), 'B': (15, 4), 'C': (10, 7), 'D': (20, 4)}
    v_label0 = dict((v, k) for (k, v) in c.items())

    edges = [('A', 'B'), ('A', 'C'), ('B', 'D'), ('C', 'B')]
    edges2 = [(c[u], c[v]) for (u, v) in edges]

    # Descomenta una de los dos lineas siguientes:
    # g0: UndirectedGraph[Vertex] = UndirectedGraph(E=edges2)
    g0: Digraph[Vertex] = Digraph(E=edges2)

    viewer = Graph2dViewer(g0, v_label0, canvas_width=400, canvas_height=300)
    viewer.run()



from __future__ import annotations

from collections.abc import Iterable, Iterator
from dataclasses import dataclass

from algoritmia.schemes.bt_scheme import DecisionSequence, bt_solve
from algoritmia.schemes.bt_scheme import ScoredDecisionSequence, bt_min_solve
from algoritmia.schemes.bt_scheme import StateDecisionSequence, bt_vc_solve

Decision = int
Score = int

# 'bt_solve' y 'bt_vc_solve' devuelven un Iterator del tipo devuelto por el método 'solution' de
# la clase 'DecisionSequence', cuya implementación por defecto devuelve una tupla con las decisiones:
SolutionDS = tuple[Decision, ...]

# 'bt_min_solve' y 'bt_max_solve' devuelven un Iterator del tipo devuelto por el método 'solution' de
# la clase 'ScoredDecisionSequence', cuya implementación por defecto devuelve una tupla de dos elementos:
# el Score y una tupla con las decisones:
SolutionSDS = tuple[Score, tuple[Decision, ...]]


def coin_change_solve_naif(v: tuple[int, ...], Q: int) -> Iterator[SolutionDS]:
    def calc(ds: tuple[int, ...]) -> int:
        return sum(ds[i] * v[i] for i in range(len(ds)))

    class CoinChangeDS(DecisionSequence[Decision]):
        def is_solution(self) -> bool:
            q = calc(self.decisions())
            return len(self) == len(v) and q == Q

        def successors(self) -> Iterable[CoinChangeDS]:
            n = len(self)
            if n < len(v):
                q = calc(self.decisions())
                pending = Q - q
                for num_coins in range(pending // v[n] + 1):
                    yield self.add_decision(num_coins)

    initial_ds = CoinChangeDS()
    return bt_solve(initial_ds)


def coin_change_solve(v: tuple[int, ...], Q: int) -> Iterator[SolutionDS]:
    @dataclass
    class Extra:
        pending: int

    class CoinChangeDS(DecisionSequence[Decision]):
        def is_solution(self) -> bool:
            return len(self) == len(v) and self.extra.pending == 0

        def successors(self) -> Iterable[CoinChangeDS]:
            n = len(self)
            if n < len(v):
                for num_coins in range(self.extra.pending // v[n] + 1):
                    pending2 = self.extra.pending - num_coins * v[n]
                    yield self.add_decision(num_coins, Extra(pending2))

    initial_ds = CoinChangeDS(Extra(Q))
    return bt_solve(initial_ds)


def coin_change_vc_solve(v: tuple[int, ...], Q: int) -> Iterator[SolutionDS]:
    @dataclass
    class Extra:
        pending: int

    class CoinChangeDS(StateDecisionSequence[Decision]):
        def is_solution(self) -> bool:
            return len(self) == len(v) and self.extra.pending == 0

        def successors(self) -> Iterable[CoinChangeDS]:
            n = len(self)
            if n < len(v):
                for num_coins in range(self.extra.pending // v[n] + 1):
                    pending2 = self.extra.pending - num_coins * v[n]
                    yield self.add_decision(num_coins, Extra(pending2))

        def state(self) -> tuple[int, int]:
            return len(self), self.extra.pending

    initial_ds = CoinChangeDS(Extra(Q))
    return bt_vc_solve(initial_ds)


def coin_change_opt_solve(v: tuple[int, ...], Q: int) -> Iterator[SolutionSDS]:
    @dataclass
    class Extra:
        pending: int

    class CoinChangeDS(ScoredDecisionSequence[Decision]):
        def is_solution(self) -> bool:
            return len(self) == len(v) and self.extra.pending == 0

        def successors(self) -> Iterable[CoinChangeDS]:
            n = len(self)
            if n < len(v):
                for num_coins in range(self.extra.pending // v[n] + 1):
                    pending2 = self.extra.pending - num_coins * v[n]
                    yield self.add_decision(num_coins, Extra(pending2))

        def state(self) -> tuple[int, int]:
            return len(self), self.extra.pending

        def score(self) -> int:
            return sum(self.decisions())

    initial_ds = CoinChangeDS(Extra(Q))
    return bt_min_solve(initial_ds)


# Programa principal ---------------------------------
if __name__ == "__main__":
    coins, quantity = (1, 2, 5), 7
    for solve in [coin_change_solve, coin_change_vc_solve, coin_change_opt_solve]:
        print(solve.__name__)
        for sol in solve(coins, quantity):
            print(sol)
        print()



from algoritmia.datastructures.graphs import UndirectedGraph, WeightingFunction

weighted_edges = {('Castelló', 'Sagunt'): 43, ('Sagunt', 'València'): 50}

wf = WeightingFunction(weighted_edges, symmetrical=True)
road_graph = UndirectedGraph(E=weighted_edges.keys())

print(wf('Castelló', 'Sagunt'))  # Output: 43



from enum import Enum
from typing import Optional, Union

from easypaint import EasyPaint

from algoritmia.datastructures.graphs import UndirectedGraph

Num = Union[int, float]
Vertex = tuple[Num, Num]


class Mode(Enum):
    ALL = 0
    MST = 1


class Graph2dMstViewer(EasyPaint):

    def __init__(self, g_all: UndirectedGraph[Vertex], g_mst: UndirectedGraph[Vertex],
                 canvas_width: int = 400, canvas_height: int = 300,
                 margin: int = 15, background='white', node_size: Optional[float] = None,
                 colors: dict[Vertex, str] = None):
        self.node_size = node_size
        self.margin = margin
        self.colors = colors
        if not isinstance(g_all, UndirectedGraph) or \
                any([not isinstance(p, tuple) or len(p) != 2 or
                     not isinstance(p[0], int) and not isinstance(p[0], float) or
                     not isinstance(p[1], int) and not isinstance(p[1], float) for p in g_all.V]):
            raise TypeError("The graph must be an UndirectedGraph. Vertices must be tuples of two integers or floats")

        self.g_all = g_all
        self.g_mst = g_mst
        self.mode = Mode.ALL
        self.g = self.g_all

        self.max_y = max(p[1] for p in self.g.V)
        self.max_x = max(p[0] for p in self.g.V)
        self.min_y = min(p[1] for p in self.g.V)
        self.min_x = min(p[0] for p in self.g.V)

        super().__init__()
        self.easypaint_configure(title='2D Graph MST Viewer - Press <space> to toogle',
                                 background=background,
                                 size=(canvas_width, canvas_height))

    def on_key_press(self, keysym: str):
        if keysym.lower() in ['return', 'escape']:
            self.close()
        elif keysym.lower() == 'space':
            if self.mode == Mode.ALL:
                self.g = self.g_mst
                self.mode = Mode.MST
            else:
                self.g = self.g_all
                self.mode = Mode.ALL
            self.erase()
            self.draw_graph()

    def draw_graph(self):
        w, h = self.size
        w -= self.margin
        h -= self.margin
        sizex = self.max_x - self.min_x + 1
        sizey = self.max_y - self.min_y + 1
        cell_size = min(w / sizex, h / sizey)

        m = ((w - cell_size * sizex) / 2 - self.min_x * cell_size + self.margin / 2,
             (h - cell_size * sizey) / 2 - self.min_y * cell_size + self.margin / 2)
        if self.node_size is None:
            self.node_size = (w * h / len(self.g.V)) ** 0.5 / 10  # cell_size / 8

        for u, v in self.g.E:
            self.create_line((u[0] + 0.5) * cell_size + m[0], (u[1] + 0.5) * cell_size + m[1],
                             (v[0] + 0.5) * cell_size + m[0], (v[1] + 0.5) * cell_size + m[1])

        if self.node_size > 0:
            for u in self.g.V:
                if self.colors is None:
                    fill = 'palegreen'
                elif u in self.colors:
                    fill = self.colors[u]
                else:
                    fill = 'white'
                self.create_filled_circle((u[0] + 0.5) * cell_size + m[0], (u[1] + 0.5) * cell_size + m[1],
                                          self.node_size, fill=fill)

    def main(self):
        self.draw_graph()


if __name__ == '__main__':
    from algoritmia.algorithms.mst import prim
    from algoritmia.data.iberia import iberia2d, km2d

    # g_msf = kruskal(iberia2d, km2d)
    g_msf = prim(iberia2d, km2d)

    gv = Graph2dMstViewer(iberia2d, g_msf, canvas_width=800, canvas_height=800, margin=50)
    gv.run()



from typing import Union

from algoritmia.utils import infinity

Quantity = int  # La cantidad a devolver
Decision = int  # Un número de de monedas
Score = Union[int, float]  # La puntuación es el número de monedas utilizadas o, si no hay solución, +infinito

LParams = tuple[int, Quantity]
Solution = tuple[Score, list[Decision]]


def solve(q: Quantity, v: list[int]) -> Solution:
    def L(n: int, q: Quantity) -> Score:
        if n == 0:
            return 0 if q == 0 else infinity
        if (n, q) not in mem:
            res: list[tuple[Score, LParams, Decision]] = []
            for d in range(q // v[n - 1] + 1):
                c_score = L(n - 1, q - d * v[n - 1]) + d
                previous = (n - 1, q - d * v[n - 1])
                res.append((c_score, previous, d))
            mem[n, q] = min(res)
        return mem[n, q][0]

    mem: dict[LParams, tuple[Score, LParams, Decision]] = {}
    score = L(len(v), q)
    n = len(v)
    sol = []
    while n > 0:
        _, (n, q), d = mem[n, q]
        sol.append(d)
    sol.reverse()
    return score, sol


if __name__ == '__main__':
    v: list[int] = [1, 2, 5, 10]
    q: Quantity = 11
    print("Instance:", q, v)
    print("Solution:", solve(q, v))



from collections.abc import Iterable
from typing import Union

from algoritmia.algorithms.traversers import bf_edge_traverser, dijkstra_edge_traverser, dijkstra_metric_edge_traverser
from algoritmia.datastructures.graphs import IGraph, Digraph, TVertex, TEdge, WeightingFunction
from algoritmia.utils import infinity

# Cinco algoritmos para obtener el camino más corto entre dos vertices de un grafo:
# - Tres basados en recorredores:
#   - shortest_path_unweighted_graph(): Para grafos no ponderados
#   - shortest_path_positive_weighted_graph(): Para grafos ponderados positivos (algoritmo de Dijkstra)
#   - shortest_path_metric_graph(): Para grafos métricos (algoritmo de Dijkstra modificado)
# - Dos que utilizan programación dinamica:
#   - shortest_path_acyclic_digraph(): Para digrafos ponderados acíclicos
#   - shortest_path_digraph()/bellman_ford(): Para digrafos ponderados sin ciclos negativos (algoritmo de Bellman-Ford)

# ----------------------------------------------------------------
# Tres algoritmos basados en recorredores
# ----------------------------------------------------------------

TPath = list[TVertex]  # Generic path


# Devuelve el camino más corto entre dos vértices en grafos no ponderados
# Coste temporal: O(|V| + |E|)
def shortest_path_unweighted_graph(g: IGraph[TVertex], v_source: TVertex, v_target: TVertex) -> TPath:
    edges = bf_edge_traverser(g, v_source)
    return path_recover(edges, v_target)


# Devuelve el camino más corto entre dos vértices en grafos ponderados positivos (algoritmo de Dijkstra)
# Coste temporal: O(|V|^2)
def shortest_path_positive_weighted_graph(g: IGraph[TVertex], d: WeightingFunction,
                                          v_source: TVertex, v_target: TVertex) -> TPath:
    edges = dijkstra_edge_traverser(g, d, v_source)  # Con diccionario: O(|V|^2)
    # edges = dijkstra_hm_edge_traverser(g, d, v_source)  # Con diccionario de prioridad: O(|V| + |E| log |V|)
    return path_recover(edges, v_target)


# Devuelve el camino más corto entre dos vértices en grafos métricos (algoritmo de Dijkstra modificado)
# Coste temporal: O(|V|^2)
# En la práctica es O(|V|) para grafos densos y O(sqrt(|V|)) para grafos dispersos.
def shortest_path_metric_graph(g: IGraph[TVertex], d: WeightingFunction,
                               v_source: TVertex, v_target: TVertex) -> TPath:
    edges = dijkstra_metric_edge_traverser(g, d, v_source, v_target)
    return path_recover(edges, v_target)


# Dada la lista de aristas que devuelve un recorredor y un vértice final,
# devuelve el camino desde el vértice inicial al final
# Coste temporal: O(|V|), porque los recorredores devuelven una arista por vértice: len(edges) = |V|
def path_recover(edges: Iterable[TEdge], v: TVertex) -> TPath:
    # Creates backpointer dictionary (bp)
    bp = {}
    for o, d in edges:
        bp[d] = o
        if d == v:  # I have all I need
            break
    # Recover the path jumping back
    path = [v]
    while v != bp[v]:
        v = bp[v]
        path.append(v)
    # reverse the path
    path.reverse()
    return path


# ----------------------------------------------------------------
# Dos algoritmos de programación dinámica
# ----------------------------------------------------------------

Score = Union[int, float]
Decision = TVertex
Solution = tuple[Score, list[Decision]]
LParams = TVertex               # Para sp_acyclic_digraph()
LParams2 = tuple[TVertex, int]  # Para sp_digraph()/bellman_ford()


# Devuelve el camino más corto entre dos vértices en digrafos ponderados acíclicos
# Coste temporal: O(|V| + |E|)
def shortest_path_acyclic_digraph(g: Digraph[TVertex],
                                  d: WeightingFunction[TVertex],
                                  v_initial: TVertex,
                                  v_final: TVertex) -> Solution:
    def L(v: TVertex) -> Score:
        if v == v_initial:
            return 0
        if v not in mem:
            res: list[tuple[Score, LParams]] = []
            for u in g.preds(v):
                c_score: Score = L(u) + d(u, v)
                previous: LParams = u
                # La decision, u, coincide con el previous
                res.append((c_score, previous))
            mem[v] = min(res, default=infinity)
        return mem[v][0]

    mem: dict[LParams, tuple[Score, LParams]] = {}
    score = L(v_final)
    if score == infinity:
        return score, []  # El vértice destino es inalcanzable desde el origen
    v0 = v_final
    sol: list[Decision] = [v0]
    while v0 != v_initial:
        _, v0 = mem[v0]
        sol.append(v0)
    sol.reverse()
    return score, sol


# Devuelve el camino más corto entre dos vértices en digrafos ponderados sin ciclos negativos
# Utiliza el algoritmo de Bellman-Ford (algoritmo de programación dinámica)
# Coste temporal: O(|V| |E|)
def shortest_path_digraph(g: Digraph[TVertex],
                          d: WeightingFunction[TVertex],
                          v_initial: TVertex,
                          v_final: TVertex) -> Solution:
    def L(v: TVertex, n: int) -> Score:
        if v == v_initial:
            return 0
        if n == 0:
            return infinity
        if (v, n) not in mem:
            res: list[tuple[Score, LParams2]] = []
            for u in g.preds(v):
                c_score: Score = L(u, n - 1) + d(u, v)
                previous: LParams2 = u, n - 1
                # La decision, u, se puede extraer del previous
                res.append((c_score, previous))
            mem[v, n] = min(res, default=infinity)
        return mem[v, n][0]

    mem: dict[LParams2, tuple[Score, LParams2]] = {}
    score: Score = L(v_final, len(g.V) - 1)
    if score == infinity:
        return score, []  # El vértice destino es inalcanzable desde el origen
    v0 = v_final
    n0 = len(g.V) - 1
    sol: list[Decision] = [v0]
    while v0 != v_initial:
        _, (v0, n0) = mem[v0, n0]
        sol.append(v0)
    sol.reverse()
    return score, sol


bellman_ford = shortest_path_digraph  # Creamos un alias para la función


# ----------------------------------------------------------------
# Programa principal
# ----------------------------------------------------------------

if __name__ == '__main__':
    # -------------
    # Output: [(0, 0), (0, 1), (1, 1), (1, 2), (0, 2), (0, 3), (1, 3)]

    from algoritmia.datastructures.graphs import UndirectedGraph

    Vertex = tuple[int, int]
    Edge = tuple[Vertex, Vertex]

    my_edges: list[Edge] = [((0, 0), (0, 1)), ((0, 2), (0, 3)), ((1, 0), (1, 1)), ((1, 1), (1, 2)),
                            ((2, 0), (2, 1)), ((2, 1), (2, 2)), ((2, 2), (2, 3)), ((0, 1), (1, 1)),
                            ((0, 2), (1, 2)), ((0, 3), (1, 3)), ((1, 1), (2, 1)), ((1, 2), (2, 2))]

    my_graph = UndirectedGraph(E=my_edges)
    v_initial0 = (0, 0)
    v_final0 = (1, 3)

    path0 = shortest_path_unweighted_graph(my_graph, v_initial0, v_final0)
    print('shortest_path_unweighted_graph:', path0)

    # ---------------
    # Output: ['Madrid', 'Venturada', 'Aranda de Duero', 'Lerma', 'Burgos', 'Miranda del Ebro', 'Armiñón', 'Basauri', 'Bilbao']

    from algoritmia.data.iberia import iberia, km

    spath = shortest_path_positive_weighted_graph(iberia, km, 'Madrid', 'Bilbao')
    print('shortest_path_positive_weighted_graph:', spath)

    # ---------------
    # Output: ['Madrid', 'Venturada', 'Aranda de Duero', 'Lerma', 'Burgos', 'Miranda del Ebro', 'Armiñón', 'Basauri', 'Bilbao']

    from algoritmia.data.iberia import iberia2d, km2d, coords2d, coords2d_inv

    spath_mg = shortest_path_metric_graph(iberia2d, km2d, coords2d['Madrid'], coords2d['Bilbao'])
    psath = [coords2d_inv[c] for c in spath_mg]
    print('shortest_path_metric_graph:', spath)

    # Digraph without cycles ----------------------
    Vertex = int
    Edge = tuple[int, int]
    Weight = int

    d0: dict[Edge, Weight] = {(0, 1): 1, (0, 3): 50, (1, 2): 10, (2, 3): 3, (3, 4): 4}
    g0 = Digraph(E=d0.keys())
    wf0 = WeightingFunction((e for e in d0.items()))
    print('shortest_path_acyclic_digraph:', shortest_path_acyclic_digraph(g0, wf0, 0, 3))

    # DiGraph with cycles ----------------------
    d1: dict[Edge, Weight] = {(0, 1): 1, (0, 3): 50, (1, 2): 10, (2, 0): 2, (2, 3): -3, (3, 1): -2, (3, 2): 100,
                              (3, 4): 4}
    g1 = Digraph(E=d1.keys())
    wf1 = WeightingFunction((e for e in d1.items()))
    print('shortest_path_digraph:', shortest_path_digraph(g1, wf1, 3, 2))



# coding: latin1

import unittest
from random import seed, shuffle

from algoritmia.datastructures.prioritymaps import (MinHeapMap, MaxHeapMap, MinFibonacciHeap, MaxFibonacciHeap)


class TestHeapMaps(unittest.TestCase):
    def setUp(self):
        self.pd1 = MinHeapMap()
        self.pairs = [('a', 1), ('z', 10), ('b', 5), ('c', 8), ('d', 12), ('e', 10)]
        self.pd2 = MinHeapMap(self.pairs)
        self.pairdict = dict(self.pairs)
        self.pd3 = MinHeapMap(self.pairdict)

    def test_ctor(self):
        self.assertEqual(len(self.pd1), 0)
        self.assertEqual(len(self.pd2), 6)
        self.assertEqual(len(self.pd2), 6)

    def test_opt(self):
        self.assertRaises(IndexError, self.pd1.opt)
        for i in (1, 5, 8, 10, 10, 12):
            self.assertEqual(self.pd2[self.pd2.opt()], i)
            self.pd2.extract_opt()
            self.assertEqual(self.pd3[self.pd3.opt()], i)
            self.pd3.extract_opt()

    def test_getitem(self):
        for k, v in self.pairs:
            self.assertEqual(self.pd2[k], v)
            self.assertEqual(self.pd3[k], v)
        self.assertRaises(KeyError, self.pd2.__getitem__, 'xx')
        self.assertRaises(KeyError, self.pd3.__getitem__, 'xx')

    def test_setitem(self):
        for pd in self.pd1, self.pd2, self.pd3:
            for k, v in pd.items():
                pd[k] = pd[k] + 1
                self.assertEqual(pd[k], v + 1)
                pd[k] = pd[k] - 10
                self.assertEqual(pd[k], v - 9)
            for k, v in ('xx', 10), ('yy', 100):
                pd[k] = v
                self.assertEqual(pd[k], v)

    def test_opt_value_and_item(self):
        for pd in self.pd1, self.pd2, self.pd3:
            if len(pd) > 0:
                self.assertEqual(pd.opt_item()[1], pd.opt_value())
                self.assertEqual(pd.opt_item()[0], pd.opt())
            else:
                self.assertRaises(IndexError, pd.opt_item)

    def test_contains(self):
        for pd in self.pd2, self.pd3:
            for k, v in self.pairs:
                self.assertTrue(k in pd)
        for pd in [self.pd1]:
            for k, v in self.pairs:
                self.assertFalse(k in pd)

    def test_del(self):
        for pd in self.pd2, self.pd3:
            for k, v in self.pairs:
                del pd[k]
            self.assertEqual(len(pd), 0)

    def test_repr(self):
        i = 0
        for pd in self.pd1, self.pd2, self.pd3:
            i += 1
            if len(pd) > 0:
                self.assertEqual(list(sorted(self.pairs)), list(sorted(eval(repr(pd)).items())))

    def test_min_and_max_prioritydics(self):
        a = MinHeapMap(self.pairs)
        b = MaxHeapMap(self.pairs)
        self.assertEqual(len(self.pairs), len(a))
        self.assertEqual(len(self.pairs), len(b))


class TestHeapMap2(unittest.TestCase):
    def setUp(self):
        seed(0)
        self.ll = [(chr(ord('a') + i), i) for i in range(26)]
        shuffle(self.ll)
        self.minpd = MinHeapMap(data=self.ll)
        self.maxpd = MaxHeapMap(data=self.ll)

    def test_opt_onNonEmptyPriorityDict_getsOptimumKey(self):
        self.assertEqual(self.minpd.opt(), 'a')
        self.assertEqual(self.maxpd.opt(), 'z')

    def test_opt_item_onNonEmptyPriorityDict_getsOptimumItem(self):
        self.assertEqual(self.minpd.opt_item(), ('a', 0))
        self.assertEqual(self.maxpd.opt_item(), ('z', 25))

    def test_opt_value_onNonEmptyPriorityDict_getsOptimumValue(self):
        self.assertEqual(self.minpd.opt_value(), 0)
        self.assertEqual(self.maxpd.opt_value(), 25)

    def test_extract_opt_onNonEmptyPriorityDict_extractsOptimumAndGetsKey(self):
        self.assertEqual(self.minpd.extract_opt(), 'a')
        self.assertEqual(self.minpd.opt(), 'b')
        self.assertEqual(self.maxpd.extract_opt(), 'z')
        self.assertEqual(self.maxpd.opt(), 'y')

    def test_extract_opt_onNonEmptyPriorityDict_extractsOptimumAndGetsItem(self):
        self.assertEqual(self.minpd.extract_opt_item(), ('a', 0))
        self.assertEqual(self.minpd.opt(), 'b')
        self.assertEqual(self.maxpd.extract_opt_item(), ('z', 25))
        self.assertEqual(self.maxpd.opt(), 'y')

    def test_contains_KeyInDict_returnsTrue(self):
        for i in range(26):
            self.assertTrue(chr(ord('a') + i) in self.minpd)
            self.assertTrue(chr(ord('a') + i) in self.maxpd)

    def test_contains_KeyNotInDict_returnsFalse(self):
        for c in 'A', '1':
            self.assertFalse(c in self.minpd)
            self.assertFalse(c in self.maxpd)

    def test_getItem_KeyInDict_returnsAssociateValue(self):
        for i in range(26):
            self.assertEqual(self.minpd[chr(ord('a') + i)], i)
            self.assertEqual(self.maxpd[chr(ord('a') + i)], i)

    def test_getItem_KeyNotInDict_raisesException(self):
        for c in 'A', '1':
            self.assertRaises(KeyError, self.minpd.__getitem__, c)
            self.assertRaises(KeyError, self.maxpd.__getitem__, c)

    def test_setItem_KeyNotInDict_createsNewItem(self):
        self.minpd['A'] = -1
        self.assertEqual(self.minpd['A'], -1)
        self.minpd['1'] = 100
        self.assertEqual(self.minpd['1'], 100)

    def test_setItem_KeyInDict_changesItem(self):
        self.minpd['a'] = -1
        self.assertEqual(self.minpd['a'], -1)
        self.minpd['z'] = 100
        self.assertEqual(self.minpd['z'], 100)

    def test_setItem_definingNewOpt_changesOpt(self):
        self.minpd['m'] = -1
        self.assertEqual(self.minpd.opt(), 'm')
        self.maxpd['m'] = 100
        self.assertEqual(self.maxpd.opt(), 'm')

    def test_delItem_existingKey_removesItem(self):
        del self.minpd['m']
        self.assertRaises(KeyError, self.minpd.__getitem__, 'm')
        del self.maxpd['m']
        self.assertRaises(KeyError, self.maxpd.__getitem__, 'm')

    def test_delItem_nonExistingKey_raisesException(self):
        self.assertRaises(KeyError, self.minpd.__delitem__, 'A')
        self.assertRaises(KeyError, self.maxpd.__delitem__, 'A')

    def test_keys_onPriorityDict_iteratesAllKeys(self):
        self.assertEqual(set(self.minpd.keys()), set((a for (a, b) in self.ll)))
        self.assertEqual(set(self.maxpd.keys()), set((a for (a, b) in self.ll)))

    def test_values_onPriorityDict_iteratesAllValues(self):
        self.assertEqual(set(self.minpd.values()), set((b for (a, b) in self.ll)))
        self.assertEqual(set(self.maxpd.values()), set((b for (a, b) in self.ll)))

    def test_items_onPriorityDict_iteratesAllItems(self):
        self.assertEqual(set(self.minpd.items()), set(self.ll))
        self.assertEqual(set(self.maxpd.items()), set(self.ll))

    def test_get_onExistingKey_returnsValue(self):
        self.assertEqual(self.minpd.get('a', 100), 0)
        self.assertEqual(self.maxpd.get('a', 100), 0)

    def test_get_onNonExistingKey_returnsDefaultValue(self):
        self.assertEqual(self.minpd.get('A', 100), 100)
        self.assertEqual(self.maxpd.get('A', 100), 100)

    def test_setdefault_onExistingKey_returnsValue(self):
        self.assertEqual(self.minpd.setdefault('a', 100), 0)
        self.assertEqual(self.maxpd.setdefault('a', 100), 0)

    def test_setdefault_onNonExistingKey_returnsNewValue(self):
        self.assertEqual(self.minpd.setdefault('A', 100), 100)
        self.assertEqual(self.maxpd.setdefault('A', 100), 100)

    def test_iter_onPriorityMap_returnsKeys(self):
        self.assertEqual(set(self.minpd), set((a for (a, b) in self.ll)))
        self.assertEqual(set(self.maxpd), set((a for (a, b) in self.ll)))

    def test_len_onPriorityMap_returnsSize(self):
        self.assertEqual(len(self.minpd), 26)
        self.assertEqual(len(self.maxpd), 26)

    def test_repr_onPriorityMap_shouldReturnEvaluableExpression(self):
        self.assertEqual(list(sorted(self.minpd)), list(sorted(eval(repr(self.minpd)))))
        self.assertEqual(list(sorted(self.maxpd)), list(sorted(eval(repr(self.maxpd)))))


class TestFibonacciHeap(unittest.TestCase):
    def setUp(self):
        self.pairs = [('a', 1), ('z', 11), ('b', 5), ('c', 8), ('d', 12), ('e', 10)]
        self.a = MinFibonacciHeap(self.pairs)

    def test_len(self):
        self.assertEqual(len(self.pairs), len(self.a))

    def test_opt(self):
        for i in (1, 5, 8, 10, 11, 12):
            self.assertEqual(self.a[self.a.opt()], i)
            self.a.extract_opt()

    def test_getitem(self):
        for k, v in self.pairs:
            self.assertEqual(self.a[k], v)
        self.assertRaises(KeyError, self.a.__getitem__, 'xx')

    def test_setitem(self):
        for k, v in self.a.items():
            v = self.a[k]
            self.assertRaises(ValueError, self.a.__setitem__, k, v + 1)
            self.a[k] = self.a[k] - 10
            self.assertEqual(self.a[k], v - 10)
        for k, v in ('xx', 10), ('yy', 100):
            self.a[k] = v
            self.assertEqual(self.a[k], v)

        v = -100
        while len(self.a) > 0:
            w = self.a.opt_value()
            self.assertTrue(v <= w)
            self.a.extract_opt()
            v = w

    def test_opt_value_and_item(self):
        if len(self.a) > 0:
            self.assertEqual(self.a.opt_item()[1], self.a.opt_value())
            self.assertEqual(self.a.opt_item()[0], self.a.opt())
        else:
            self.assertRaises(IndexError, self.a.opt_item)

    def test_contains(self):
        for k, v in self.pairs:
            self.assertTrue(k in self.a)
        for k in 'x', 'y', '':
            self.assertFalse(k in self.a)

    def test_del(self):
        for k, v in self.pairs:
            del self.a[k]
        self.assertEqual(len(self.a), 0)

    def test_repr(self):
        self.assertEqual(len(self.a), len(eval(repr(self.a))))


class TestFibonacciHeap2(unittest.TestCase):
    def setUp(self):
        seed(0)
        self.ll = [(chr(ord('a') + i), i) for i in range(26)]
        shuffle(self.ll)
        self.minpd = MinFibonacciHeap(data=self.ll)
        self.maxpd = MaxFibonacciHeap(data=self.ll)

    def test_opt_onNonEmptyPriorityDict_getsOptimumKey(self):
        self.assertEqual(self.minpd.opt(), 'a')
        self.assertEqual(self.maxpd.opt(), 'z')

    def test_opt_item_onNonEmptyPriorityDict_getsOptimumItem(self):
        self.assertEqual(self.minpd.opt_item(), ('a', 0))
        self.assertEqual(self.maxpd.opt_item(), ('z', 25))

    def test_opt_value_onNonEmptyPriorityDict_getsOptimumValue(self):
        self.assertEqual(self.minpd.opt_value(), 0)
        self.assertEqual(self.maxpd.opt_value(), 25)

    def test_extract_opt_onNonEmptyPriorityDict_extractsOptimumAndGetsKey(self):
        self.assertEqual(self.minpd.extract_opt(), 'a')
        self.assertEqual(self.minpd.opt(), 'b')
        self.assertEqual(self.maxpd.extract_opt(), 'z')
        self.assertEqual(self.maxpd.opt(), 'y')

    def test_extract_opt_onNonEmptyPriorityDict_extractsOptimumAndGetsItem(self):
        self.assertEqual(self.minpd.extract_opt_item(), ('a', 0))
        self.assertEqual(self.minpd.opt(), 'b')
        self.assertEqual(self.maxpd.extract_opt_item(), ('z', 25))
        self.assertEqual(self.maxpd.opt(), 'y')

    def test_contains_KeyInDict_returnsTrue(self):
        for i in range(26):
            self.assertTrue(chr(ord('a') + i) in self.minpd)
            self.assertTrue(chr(ord('a') + i) in self.maxpd)

    def test_contains_KeyNotInDict_returnsFalse(self):
        for c in 'A', '1':
            self.assertFalse(c in self.minpd)
            self.assertFalse(c in self.maxpd)

    def test_getItem_KeyInDict_returnsAssociateValue(self):
        for i in range(26):
            self.assertEqual(self.minpd[chr(ord('a') + i)], i)
            self.assertEqual(self.maxpd[chr(ord('a') + i)], i)

    def test_getItem_KeyNotInDict_raisesException(self):
        for c in 'A', '1':
            self.assertRaises(KeyError, self.minpd.__getitem__, c)
            self.assertRaises(KeyError, self.maxpd.__getitem__, c)

    def test_setItem_KeyNotInDict_createsNewItem(self):
        self.minpd['A'] = -1
        self.assertEqual(self.minpd['A'], -1)
        self.minpd['1'] = 100
        self.assertEqual(self.minpd['1'], 100)

    def test_setItem_KeyInDict_changesItem(self):
        self.minpd['a'] = -1
        self.assertEqual(self.minpd['a'], -1)
        self.maxpd['z'] = 100
        self.assertEqual(self.maxpd['z'], 100)

    def test_setItem_KeyInDictButNotImproving_raisesException(self):
        self.assertRaises(ValueError, self.minpd.__setitem__, 'a', 100)
        self.assertRaises(ValueError, self.maxpd.__setitem__, 'a', -1)

    def test_setItem_definingNewOpt_changesOpt(self):
        self.minpd['m'] = -1
        self.assertEqual(self.minpd.opt(), 'm')
        self.maxpd['m'] = 100
        self.assertEqual(self.maxpd.opt(), 'm')

    def test_delItem_existingKey_removesItem(self):
        del self.minpd['m']
        self.assertRaises(KeyError, self.minpd.__getitem__, 'm')
        del self.maxpd['m']
        self.assertRaises(KeyError, self.maxpd.__getitem__, 'm')

    def test_delItem_nonExistingKey_raisesException(self):
        self.assertRaises(KeyError, self.minpd.__delitem__, 'A')
        self.assertRaises(KeyError, self.maxpd.__delitem__, 'A')

    def test_keys_onPriorityDict_iteratesAllKeys(self):
        self.assertEqual(set(self.minpd.keys()), set((a for (a, b) in self.ll)))
        self.assertEqual(set(self.maxpd.keys()), set((a for (a, b) in self.ll)))

    def test_values_onPriorityDict_iteratesAllValues(self):
        self.assertEqual(set(self.minpd.values()), set((b for (a, b) in self.ll)))
        self.assertEqual(set(self.maxpd.values()), set((b for (a, b) in self.ll)))

    def test_items_onPriorityDict_iteratesAllItems(self):
        self.assertEqual(set(self.minpd.items()), set(self.ll))
        self.assertEqual(set(self.maxpd.items()), set(self.ll))

    def test_get_onExistingKey_returnsValue(self):
        self.assertEqual(self.minpd.get('a', 100), 0)
        self.assertEqual(self.maxpd.get('a', 100), 0)

    def test_get_onNonExistingKey_returnsDefaultValue(self):
        self.assertEqual(self.minpd.get('A', 100), 100)
        self.assertEqual(self.maxpd.get('A', 100), 100)

    def test_setdefault_onExistingKey_returnsValue(self):
        self.assertEqual(self.minpd.setdefault('a', 100), 0)
        self.assertEqual(self.maxpd.setdefault('a', 100), 0)

    def test_setdefault_onNonExistingKey_returnsNewValue(self):
        self.assertEqual(self.minpd.setdefault('A', 100), 100)
        self.assertEqual(self.maxpd.setdefault('A', 100), 100)

    def test_iter_onPriorityMap_returnsKeys(self):
        self.assertEqual(set(self.minpd), set((a for (a, b) in self.ll)))
        self.assertEqual(set(self.maxpd), set((a for (a, b) in self.ll)))

    def test_len_onPriorityMap_returnsSize(self):
        self.assertEqual(len(self.minpd), 26)
        self.assertEqual(len(self.maxpd), 26)

    def test_repr_onPriorityMap_shouldReturnEvaluableExpression(self):
        self.assertEqual(list(sorted(self.minpd)), list(sorted(eval(repr(self.minpd)))))
        self.assertEqual(list(sorted(self.maxpd)), list(sorted(eval(repr(self.maxpd)))))


if __name__ == "__main__":
    # import sys;sys.argv = ['', 'Test.testName']
    unittest.main()



from abc import abstractmethod, ABC
from collections.abc import Iterable, Iterator, Sized, Collection
from typing import TypeVar, Generic

T = TypeVar('T')


class IMergeFindSet(ABC, Generic[T], Iterable, Sized):
    @abstractmethod
    def add(self, x: T): pass

    @abstractmethod
    def find(self, x: T) -> T: pass

    @abstractmethod
    def merge(self, x: T, y: T): pass


class MergeFindSet(IMergeFindSet[T]):
    def __init__(self, sets: Iterable[Collection[T]] = ()):
        self._parent: dict[T, T] = {}
        self._rank: dict[T, int] = {}
        self._length = 0
        first = None
        for s in sets:
            if len(s) > 0:
                self._length += 1
                for item in s:
                    if first is None:
                        first = item
                    self._parent[item] = first
                    self._rank[item] = self._rank.get(item, 0) + 1
                first = None

    def add(self, x: T):
        self._parent[x] = x
        self._rank[x] = 1
        self._length += 1

    def merge(self, x: T, y: T):
        u = self.find(x)
        v = self.find(y)
        if u != v:
            self._length -= 1
            if self._rank[u] < self._rank[v]:
                self._parent[u] = v
            elif self._rank[u] > self._rank[v]:
                self._parent[v] = u
            else:
                self._parent[v] = u
                self._rank[u] += 1

    def find(self, x: T) -> T:
        r = x
        while r != self._parent[r]:
            r = self._parent[r]
        while x != self._parent[x]:
            self._parent[x], x = r, self._parent[x]
        return r

    def __iter__(self) -> Iterator[Iterable[T]]:
        aux = {}
        for key in self._parent:
            aux.setdefault(self.find(key), []).append(key)
        for s in aux.values():
            yield s

    def __len__(self) -> int:
        return self._length

    def __repr__(self) -> str:
        return '{}({!r})'.format(self.__class__.__name__, tuple(self))



"""
Version: 4.1 (29-sep-2022)
         4.0 (23-oct-2021)

@author: David Llorens (dllorens@uji.es)
         (c) Universitat Jaume I 2022
@license: GPL3
"""
from __future__ import annotations

from abc import abstractmethod
from functools import total_ordering
from typing import Optional, Generic, Union, Any

from algoritmia.datastructures.priorityqueues import MaxHeap, MinHeap
from algoritmia.schemes.bt_scheme import DecisionSequence, TDecision

Score = Union[int, float]

# 'bab_max_solve' y 'bab_min_solve' devuelven un Iterator del tipo devuelto por el método 'solution()' de
# la clase 'BoundedDecisionSequence', cuya implementación por defecto devuelve una tupla de dos elementos:
# el Score y una tupla con las decisones:
# tuple[Score, tuple[TDecision, ...]]
Solution = Any

@total_ordering  # Implementando  < y ==, genera el resto
class BoundedDecisionSequence(Generic[TDecision],
                              DecisionSequence[TDecision]):
    __slots__ = ("_pes", "_opt")

    def __init__(self, extra_fields=None, decisions: tuple[TDecision, ...] = ()):
        super().__init__(extra_fields, decisions)
        self._pes: Score = self.calculate_pes_bound()
        self._opt: Score = self.calculate_opt_bound()

    # Nota: Para no repetir cálculos, y dado que son inmutables, las cotas se
    # deberían calcular sólo una vez. Los métodos calculate_opt_bound u calculate_pes_bound
    # se calculan solo una vez y se guarda en self._opt y self._pes.
    # - Los métodos opt() y pes() sólo devuelven el valor de estos atributos.
    @abstractmethod
    def calculate_opt_bound(self) -> Score:
        pass

    @abstractmethod
    def calculate_pes_bound(self) -> Score:
        pass

    # Optimistic bound. Must be iqual to score() for full solutions
    def opt(self) -> Score:
        return self._opt

    # Pessimistic bound. Must be iqual to score() for full solutions
    def pes(self) -> Score:
        return self._pes

    def solution(self) -> Solution:
        return self._pes, self.decisions()

    # Comparar dos BabDecisionSequence es comparar sus cotas optimistas
    def __lt__(self, other: BoundedDecisionSequence[TDecision]) -> bool:
        return self._opt < other._opt

    def __eq__(self, other: BoundedDecisionSequence[TDecision]) -> bool:
        return self._opt == other._opt



def bab_max_solve(initial_ds: BoundedDecisionSequence[TDecision]) -> Optional[Solution]:
    heap = MaxHeap()
    heap.add(initial_ds)
    bps = initial_ds.pes()
    while len(heap) > 0:
        best_ps = heap.extract_opt()
        if best_ps.is_solution():
            return best_ps.solution()
        for new_ps in best_ps.successors():
            bps = max(bps, new_ps.pes())
            if new_ps.opt() >= bps:
                heap.add(new_ps)


def bab_min_solve(initial_ds: BoundedDecisionSequence[TDecision]) -> Optional[Solution]:
    heap = MinHeap()
    heap.add(initial_ds)
    bps = initial_ds.pes()
    while len(heap) > 0:
        best_ps = heap.extract_opt()
        if best_ps.is_solution():
            return best_ps.solution()
        for new_ps in best_ps.successors():
            bps = min(bps, new_ps.pes())
            if new_ps.opt() <= bps:
                heap.add(new_ps)



from algoritmia.viewers.graph2d_viewer import Graph2dViewer
from algoritmia.algorithms.traversers import dijkstra_edge_traverser, dijkstra_metric_edge_traverser
from algoritmia.data.iberia import iberia2d, km2d, coords2d

if __name__ == '__main__':
    v_initial = coords2d["Madrid"]
    v_final = coords2d["Bilbao"]

    edges = dijkstra_edge_traverser(iberia2d, km2d, v_initial)
    # edges = dijkstra_metric_edge_traverser(iberia2d, km2d, v_initial, v_final)
    colors = {}
    for (u, v) in edges:
        colors[v] = 'red'
        if v == v_final: break
    colors[v_initial] = 'palegreen'
    colors[v_final] = 'palegreen'

    gv = Graph2dViewer(iberia2d, canvas_width=800, canvas_height=800, vertexmode=Graph2dViewer.X_Y, colors=colors)
    gv.run()



from collections.abc import Iterator, Callable
from math import sqrt

from algoritmia.datastructures.graphs import IGraph, TVertex, TEdge, WeightingFunction, Weight
from algoritmia.datastructures.prioritymaps import MinHeapMap
from algoritmia.datastructures.queues import Fifo
from algoritmia.utils import argmin, infinity

TVertexTraverser = Callable[[IGraph[TVertex], TVertex], Iterator[TVertex]]
TEdgeTraverser = Callable[[IGraph[TVertex], TVertex], Iterator[TEdge]]


# TV traversers ----------------------------------------------------------------

def bf_vertex_traverser(graph: IGraph[TVertex],
                        v_initial: TVertex) -> Iterator[TVertex]:
    queue = Fifo()
    seen = set()
    queue.push(v_initial)
    seen.add(v_initial)
    while len(queue) > 0:
        v = queue.pop()
        yield v
        for suc in graph.succs(v):
            if suc not in seen:
                queue.push(suc)
                seen.add(suc)


def df_vertex_traverser(graph: IGraph[TVertex],
                        v_initial: TVertex) -> Iterator[TVertex]:
    def traverse_from(v: TVertex) -> Iterator[TVertex]:
        seen.add(v)
        yield v  # pre-order
        for suc in graph.succs(v):
            if suc not in seen:
                yield from traverse_from(suc)
        # yield v # post-order

    seen = set()
    yield from traverse_from(v_initial)


# TE traversers ----------------------------------------------------------------

def bf_edge_traverser(graph: IGraph[TVertex],
                      v_initial: TVertex) -> Iterator[TEdge]:
    queue = Fifo()
    seen = set()
    queue.push((v_initial, v_initial))
    seen.add(v_initial)
    while len(queue) > 0:
        u, v = queue.pop()
        yield u, v
        for suc in graph.succs(v):
            if suc not in seen:
                queue.push((v, suc))
                seen.add(suc)


def df_edge_traverser(graph: IGraph[TVertex],
                      v_initial: TVertex) -> Iterator[TEdge]:
    def traverse_from(u: TVertex, v: TVertex) -> Iterator[TEdge]:
        seen.add(v)
        yield u, v  # pre-order
        for suc in graph.succs(v):
            if suc not in seen:
                yield from traverse_from(v, suc)
        # yield u, v # post-order

    seen = set()
    return traverse_from(v_initial, v_initial)


# Con diccionario: O(|V|^2)
def dijkstra_edge_traverser(g: IGraph[TVertex],
                            d: WeightingFunction[TVertex],
                            v_initial: TVertex) -> Iterator[TEdge]:
    D: dict[TVertex, Weight] = dict((v, infinity) for v in g.V)
    D[v_initial] = 0
    bp: dict[TVertex, TVertex] = {v_initial: v_initial}
    fixed: set[TVertex] = set()
    while len(bp) > 0:  # O(|V|) veces
        v = argmin(bp.keys(), lambda u: D[u])  # O(|V|)
        fixed.add(v)
        pred_v = bp[v]
        yield pred_v, v
        del bp[v]
        for suc_v in g.succs(v):  # O(|V|)
            if suc_v not in fixed and D[v] + d(v, suc_v) < D[suc_v]:
                D[suc_v] = D[v] + d(v, suc_v)
                bp[suc_v] = v


# Con diccionario de prioridad: O(|V| + |E| log |V|)
def dijkstra_hm_edge_traverser(g: IGraph[TVertex],
                               d: WeightingFunction[TVertex],
                               v_initial: TVertex) -> Iterator[TEdge]:
    D: MinHeapMap[TVertex, Weight] = MinHeapMap((v, infinity) for v in g.V)  # O(|V|)
    D[v_initial] = 0
    bp: dict[TVertex, TVertex] = {v_initial: v_initial}
    fixed: set[TVertex] = set()
    while len(D) > 0:
        v, dv = D.extract_opt_item()  # O(log |V|), O(|V|) veces
        fixed.add(v)
        pred_v = bp[v]
        yield pred_v, v
        for suc_v in g.succs(v):
            if suc_v not in fixed and dv + d(v, suc_v) < D[suc_v]:
                D[suc_v] = dv + d(v, suc_v)  # O(log |V|), O(|E|) veces
                bp[suc_v] = v


# Distancia euclídea
def d_eu(u, v):
    a = u[0] - v[0]
    b = u[1] - v[1]
    return sqrt(a * a + b * b)


# Con diccionario: O(|V|^2)
# En la práctica el coste es O(|V|) para grafos densos y O(sqrt(|V|)) para grafos dispersos.
def dijkstra_metric_edge_traverser(g: IGraph[TVertex],
                                   d: WeightingFunction[TVertex],
                                   v_initial: TVertex, v_final: TVertex) -> Iterator[TEdge]:
    D: dict[TVertex, Weight] = dict((v, infinity) for v in g.V)
    D[v_initial] = 0
    bp: dict[TVertex, TVertex] = {v_initial: v_initial}
    fixed: set[TVertex] = set()
    while len(bp) > 0:
        v = argmin(bp.keys(), lambda u: D[u] + d_eu(u, v_final))
        fixed.add(v)
        pred_v = bp[v]
        yield pred_v, v
        if v == v_final:
            break
        del bp[v]
        for suc_v in g.succs(v):
            if suc_v not in fixed and D[v] + d(v, suc_v) < D[suc_v]:
                D[suc_v] = D[v] + d(v, suc_v)
                bp[suc_v] = v


if __name__ == '__main__':
    from algoritmia.datastructures.graphs import UndirectedGraph

    Vertex = tuple[int, int]
    Edge = tuple[Vertex, Vertex]

    my_edges = [((0, 0), (0, 1)), ((0, 2), (0, 3)), ((1, 0), (1, 1)), ((0, 1), (0, 2)),
                ((2, 0), (1, 0)), ((2, 1), (2, 2)), ((2, 2), (2, 3)), ((0, 1), (1, 1)),
                ((0, 2), (1, 2)), ((0, 3), (1, 3)), ((1, 1), (2, 1)), ((1, 2), (2, 2))]

    my_graph = UndirectedGraph[Vertex](E=my_edges)
    my_vertex_source = (0, 0)

    print('vertex_traverser_breadthfirst:', list(bf_vertex_traverser(my_graph, my_vertex_source)))
    print('edge_traverser_breadthfirst', list(bf_edge_traverser(my_graph, my_vertex_source)))

    print('vertex_traverser_depthfirst', list(df_vertex_traverser(my_graph, my_vertex_source)))
    print('edge_traverser_depthfirst', list(df_edge_traverser(my_graph, my_vertex_source)))

    # ---------------

    from algoritmia.data.mallorca import Mallorca, km

    print('edge_traverser_dijkstra', list(dijkstra_edge_traverser(Mallorca, km, 'Andratx')))



from algoritmia.datastructures.graphs import UndirectedGraph, TVertex


# agrupa los vertices que pueden compartir color
def coloring_solve(g: UndirectedGraph[TVertex]) -> list[set[TVertex]]:
    solution: list[set[TVertex]] = []
    vertex_set = set(g.V)  # Creamos una copia de los vértices

    while len(vertex_set) > 0:
        grupo = set()  # Creamos un nuevo grupo vacío
        # Añadimos los vértices cuyos sucesores no estén ya en grupo
        for v in vertex_set:
            # Simple code --------
            #can_add_v = True
            #for suc in g.succs(v):
            #    if suc in grupo:
            #        can_add_v = False  # v tiene un sucesor en el grupo
            #        break
            #if can_add_v:
            #    grupo.add(v)

            # Estas dos lineas (con generadores) reemplazan a las siete anteriores
            if not any(suc in grupo for suc in g.succs(v)):
                grupo.add(v)

        solution.append(grupo)  # Añadimos el grupo a la solución
        vertex_set -= grupo     # Quitamos los vértices del grupo de V

    return solution


if __name__ == '__main__':
    graph = UndirectedGraph(E=[(0, 1), (0, 2), (0, 3), (1, 3), (1, 4), (2, 3), (2, 5),
                               (3, 4), (3, 5), (3, 6), (4, 7), (5, 6), (5, 8), (6, 7),
                               (6, 8), (6, 9), (7, 9), (8, 9)])
    print(coloring_solve(graph))



from collections.abc import Callable, Iterable
from typing import Optional, TypeVar

infinity = float("+inf")

T = TypeVar('T')
S = TypeVar('S')


def argmin(iterable: Iterable[T], fn: Callable[[T], S], ifempty: Optional[T] = None) -> T:
    try:
        return min((fn(x), x) for x in iterable)[1]
    except ValueError:
        return ifempty


def argmax(iterable: Iterable[T], fn: Callable[[T], S], ifempty: Optional[T] = None) -> T:
    try:
        return max((fn(x), x) for x in iterable)[1]
    except ValueError:
        return ifempty



from abc import abstractmethod
from collections.abc import Sequence, Iterable, Iterator
from typing import Union, Optional

from algoritmia.datastructures.prioritymaps import IPriorityMap, K, T


class IDoubleEndedPriorityDict(IPriorityMap[K, T]):
    @abstractmethod
    def worst(self) -> K: pass

    @abstractmethod
    def worst_value(self) -> T: pass

    @abstractmethod
    def worst_item(self) -> tuple[K, T]: pass

    @abstractmethod
    def extract_worst(self) -> K: pass

    @abstractmethod
    def extract_worst_item(self) -> tuple[K, T]: pass


class MinMaxIntervalHeapMap(IDoubleEndedPriorityDict[K, T]):
    def __init__(self, data: Union[Iterable[tuple[K, T]], dict[K, T]] = (), capacity: int = 0):
        super().__init__()
        if isinstance(data, dict):
            data = data.items()
        elif not isinstance(data, Sequence):
            data = tuple(data)

        self._size = len(data)
        capacity = max(capacity, self._size)
        self._heap: list[Optional[tuple[K, T]]] = [(v, k) for (k, v) in data]
        self._heap.extend([None] * (capacity - self._size))
        self._index = {}
        for i in range(self._size): self._index[self._heap[i][1]] = i
        for v in range(0, self._size, 2):
            if v + 1 < self._size: self._swap(v)
        last_parent = self._parent(self._size - 1)
        for v in range(last_parent, -1, -2):
            self._heapify_min(v)
            self._heapify_max(v)

    def _children(self, i: int) -> Iterator[int]:
        j = 2 * (i + 1)
        if j < self._size: yield j
        j += 2
        if j < self._size: yield j

    @staticmethod
    def _parent(i: int) -> int:
        return ((i // 2 - 1) // 2) * 2

    def _swap(self, i: int) -> bool:
        if self._heap[i][0] > self._heap[i + 1][0]:
            self._heap[i], self._heap[i + 1] = self._heap[i + 1], self._heap[i]
            self._index[self._heap[i][1]] = i
            self._index[self._heap[i + 1][1]] = i + 1
            return True
        return False

    def __getitem__(self, key: K) -> T:
        if key not in self._index: raise KeyError(key)
        return self._heap[self._index[key]][0]

    def __delitem__(self, key: K):
        if key not in self._index: raise KeyError(key)
        (newvalue, newkey) = self._heap[self._size - 1]
        self._heap[self._size - 1] = None
        self._size -= 1
        i = self._index[key]
        del self._index[key]
        if i != self._size:
            self._heap[i] = (self._heap[i][0], newkey)
            self._index[newkey] = i
            self[newkey] = newvalue

    def __setitem__(self, key: K, v: T) -> T:
        if key in self._index:
            i = self._index[key]
            ii = i if i % 2 == 0 else i - 1
            oldvalue = self._heap[i][0]
            self._heap[i] = (v, key)
            if ii + 1 < self._size:
                if self._swap(ii):
                    oldvalue = self._heap[i][0]
                    if oldvalue < v:
                        self._bubble_up_max(ii)
                        self._heapify_min(ii)
                    elif oldvalue > v:
                        self._bubble_up_min(ii)
                        self._heapify_max(ii)
                else:
                    if oldvalue < v:
                        if ii == i:
                            self._heapify_min(ii)
                        else:
                            self._bubble_up_max(ii)
                    elif oldvalue > v:
                        if ii == i:
                            self._bubble_up_min(ii)
                        else:
                            self._heapify_max(ii)
            else:
                if v < oldvalue:
                    self._bubble_up_min(ii)
                elif v > oldvalue:
                    parent = self._parent(ii)
                    if parent >= 0:
                        if self._heap[parent + 1][0] < v:
                            self._heap[parent + 1], self._heap[ii] = \
                                self._heap[ii], self._heap[parent + 1]
                            self._index[self._heap[parent + 1][1]] = parent + 1
                            self._index[self._heap[ii][1]] = ii
                            self._bubble_up_max(parent)
        else:
            # self._index = self.redimMap(self._index, key) # TODO (XXX)
            if self._size == len(self._heap):
                self._heap.append(None)
            if self._size == 0:
                self._heap[0] = (v, key)
                self._index[key] = 0
                self._size += 1
            else:
                self._heap[self._size] = (v, key)
                self._index[key] = self._size
                self._size += 1
                if self._size % 2 == 0:
                    ii = self._size - 2
                    if self._swap(ii):
                        self._bubble_up_min(ii)
                    else:
                        self._bubble_up_max(ii)
                else:
                    ii = self._size - 1
                    parent = self._parent(ii)
                    if parent >= 0:
                        if self._heap[parent][0] <= v <= self._heap[parent + 1][0]: return v
                        if v > self._heap[parent + 1][0]:
                            self._heap[parent + 1], self._heap[ii] = self._heap[ii], self._heap[parent + 1]
                            self._index[self._heap[parent + 1][1]] = parent + 1
                            self._index[self._heap[ii][1]] = ii
                            self._bubble_up_max(parent)
                        elif v < self._heap[parent][0]:
                            self._heap[parent], self._heap[ii] = self._heap[ii], self._heap[parent]
                            self._index[self._heap[parent][1]] = parent
                            self._index[self._heap[ii][1]] = ii
                            self._bubble_up_min(parent)
        return v

    def min(self) -> K:
        if self._size == 0: raise IndexError("Empty Interval Heap")
        return self._heap[0][1]

    def min_value(self) -> T:
        if self._size == 0: raise IndexError("Empty Interval Heap")
        return self._heap[0][0]

    def min_item(self) -> tuple[K, T]:
        if self._size == 0: raise IndexError("Empty Interval Heap")
        return self._heap[0][1], self._heap[0][0]

    def max(self) -> K:
        if self._size == 0: raise IndexError("Empty Interval Heap")
        if self._size == 1: return self._heap[0][1]
        return self._heap[1][1]

    def max_value(self) -> T:
        if self._size == 0: raise IndexError("Empty Interval Heap")
        if self._size == 1: return self._heap[0][0]
        return self._heap[1][0]

    def max_item(self) -> tuple[K, T]:
        if self._size == 0: raise IndexError("Empty Interval Heap")
        if self._size == 1: return self._heap[0][1], self._heap[0][0]
        return self._heap[1][1], self._heap[1][0]

    def extract_min(self) -> K:
        return self.extract_min_item()[0]

    def extract_min_value(self) -> T:
        return self.extract_min_item()[1]

    def extract_min_item(self) -> tuple[K, T]:
        if self._size == 0: raise IndexError("Empty Interval Heap")
        retval = (self._heap[0][1], self._heap[0][0])
        if self._size <= 2:
            if self._size == 1:
                del self._index[self._heap[0][1]]
                self._heap[0] = None
            else:
                del self._index[self._heap[0][1]]
                self._heap[0], self._heap[1] = self._heap[1], None
                self._index[self._heap[0][1]] = 0
            self._size -= 1
            return retval
        del self._index[self._heap[0][1]]
        self._heap[0], self._heap[self._size - 1] = self._heap[self._size - 1], None
        self._index[self._heap[0][1]] = 0
        self._size -= 1
        self._heapify_min(0)
        return retval

    def extract_max(self) -> K:
        return self.extract_max_item()[0]

    def extract_max_value(self) -> T:
        return self.extract_max_item()[1]

    def extract_max_item(self) -> tuple[K, T]:
        if self._size == 0: raise Exception("Empty Interval Heap")
        if self._size == 1:
            retval = (self._heap[0][1], self._heap[0][0])
            del self._index[self._heap[0][1]]
            self._heap[0] = None
            self._size -= 1
            return retval
        retval = (self._heap[1][1], self._heap[1][0])
        if self._size == 2:
            del self._index[self._heap[1][1]]
            self._heap[1] = None
            self._size -= 1
            return retval
        del self._index[self._heap[1][1]]
        self._heap[1], self._heap[self._size - 1] = self._heap[self._size - 1], None
        self._index[self._heap[1][1]] = 1
        self._size -= 1
        self._heapify_max(0)
        return retval

    def _heapify_min(self, i: int):
        smallest = i
        while True:
            for j in self._children(i):
                if self._heap[j][0] < self._heap[smallest][0]: smallest = j
            if smallest == i: break
            self._heap[smallest], self._heap[i] = self._heap[i], self._heap[smallest]
            self._index[self._heap[smallest][1]] = smallest
            self._index[self._heap[i][1]] = i
            i = smallest
            if i + 1 < self._size:
                self._swap(i)

    def _heapify_max(self, i: int):
        largest = i + 1
        while True:
            for j in self._children(i):
                if j + 1 < self._size:
                    if self._heap[j + 1][0] > self._heap[largest][0]: largest = j + 1
                else:
                    if self._heap[j][0] > self._heap[largest][0]: largest = j
            if largest == i + 1: break
            self._heap[largest], self._heap[i + 1] = self._heap[i + 1], self._heap[largest]
            self._index[self._heap[largest][1]] = largest
            self._index[self._heap[i + 1][1]] = i + 1
            i = largest // 2 * 2
        if i + 1 < self._size:
            self._swap(i)

    def _bubble_up_max(self, i: int):
        while True:
            parent = self._parent(i)
            if parent < 0: return
            if self._heap[parent + 1][0] >= self._heap[i + 1][0]: return
            self._heap[parent + 1], self._heap[i + 1] = self._heap[i + 1], self._heap[parent + 1]
            self._index[self._heap[parent + 1][1]] = parent + 1
            self._index[self._heap[i + 1][1]] = i + 1
            i = parent

    def _bubble_up_min(self, i: int):
        while True:
            parent = self._parent(i)
            if parent < 0: return
            if self._heap[parent][0] <= self._heap[i][0]: return
            self._heap[parent], self._heap[i] = self._heap[i], self._heap[parent]
            self._index[self._heap[parent][1]] = parent
            self._index[self._heap[i][1]] = i
            i = parent

    def keys(self) -> Iterator[K]:
        for key in self._index: yield key

    def values(self) -> Iterator[T]:
        for key in self._index: yield self._heap[self._index[key]][0]

    def items(self) -> Iterator[tuple[K, T]]:
        for key in self._index:
            yield self._heap[self._index[key]][1], self._heap[self._index[key]][0]

    def get(self, key: K, default: T = None):
        if key in self._index: return self[key]
        return default

    def setdefault(self, key: K, default: T = None):
        if key in self._index: return self[key]
        self[key] = default
        return default

    def __contains__(self, key: K) -> bool:
        return key in self._index

    def __iter__(self) -> Iterator[K]:
        for key in self._index: yield key

    def __len__(self) -> int:
        return self._size

    def __repr__(self) -> str:
        return '{}({!r})'.format(self.__class__.__name__,
                                 [(k, v) for (v, k) in self._heap[:self._size]])

    opt = min
    opt_value = min_value
    opt_item = min_item
    extract_opt = extract_min
    extract_opt_item = extract_min_item

    worst = max
    worst_value = max_value
    worst_item = max_item
    extract_worst = extract_max
    extract_worst_item = extract_max_item


class MaxMinIntervalHeapMap(MinMaxIntervalHeapMap):
    opt = MinMaxIntervalHeapMap.max
    opt_value = MinMaxIntervalHeapMap.max_value
    opt_item = MinMaxIntervalHeapMap.max_item
    extract_opt = MinMaxIntervalHeapMap.extract_max
    extract_opt_item = MinMaxIntervalHeapMap.extract_max_item

    worst = MinMaxIntervalHeapMap.min
    worst_value = MinMaxIntervalHeapMap.min_value
    worst_item = MinMaxIntervalHeapMap.min_item
    extract_worst = MinMaxIntervalHeapMap.extract_min
    extract_worst_item = MinMaxIntervalHeapMap.extract_min_item



from algoritmia.datastructures.graphs import UndirectedGraph, WeightingFunction, TVertex, TEdge
from algoritmia.datastructures.mergefindsets import MergeFindSet
from algoritmia.datastructures.prioritymaps import MinHeapMap
from algoritmia.utils import argmin


# MST para grafos no diriguidos: kruslkal y prim
# Para grafos diriguidos ver algoritmo de Chu-Liu/Edmonds

def kruskal(g: UndirectedGraph[TVertex],
            d: WeightingFunction[TVertex]) -> UndirectedGraph[TVertex]:
    edges: list[TEdge] = []
    forest: MergeFindSet[TVertex] = MergeFindSet((v,) for v in g.V)
    n = 0
    for (u, v) in sorted(g.E, key=lambda e: d(e)):
        if forest.find(u) != forest.find(v):
            forest.merge(u, v)
            edges.append((u, v))
            n += 1
            if n == len(g.V) - 1:
                break
    return UndirectedGraph(E=edges)


# Use bp: dict[TVertex, TVertex]
def prim_dic(g: UndirectedGraph[TVertex],
             d: WeightingFunction[TVertex]) -> UndirectedGraph[TVertex]:
    edges: list[TEdge] = []
    fixed: set[TVertex] = set()
    for u in g.V:
        if u not in fixed:
            fixed.add(u)
            bp = dict((v, u) for v in g.succs(u))
            while len(bp) > 0:
                (v, u) = argmin(bp.items(), d)
                del bp[v]
                fixed.add(v)
                edges.append((u, v))
                for w in g.succs(v):
                    if w in fixed: continue
                    if w not in bp or d(v, w) < d(bp[w], w):
                        bp[w] = v
    return UndirectedGraph(E=edges)


# Uses MinHeapMap[Vertex, tuple[Weight, Vertex]]
def prim(g: UndirectedGraph[TVertex],
         d: WeightingFunction[TVertex]) -> UndirectedGraph[TVertex]:
    edges: list[TEdge] = []
    fixed: set[TVertex] = set()
    for u in g.V:
        if u not in fixed:
            fixed.add(u)
            min_hm = MinHeapMap((v, (d(u, v), u)) for v in g.succs(u))
            while len(min_hm) > 0:
                (v, (_, u)) = min_hm.extract_opt_item()
                fixed.add(v)
                edges.append((u, v))
                for w in g.succs(v):
                    if w in fixed: continue
                    if w not in min_hm or d(v, w) < d(min_hm[w][1], w):
                        min_hm[w] = (d(v, w), v)
    return UndirectedGraph(E=edges)


if __name__ == '__main__':
    from algoritmia.data.mallorca import Mallorca, km

    mst_kruskal = kruskal(Mallorca, km)
    print(mst_kruskal)

    mst_prim_dic = prim_dic(Mallorca, km)
    print(mst_prim_dic)

    mst_prim = prim(Mallorca, km)
    print(mst_prim)



from algoritmia.datastructures.graphs import UndirectedGraph

edges = [('Castelló', 'Sagunt'), ('Sagunt', 'València')]
ug = UndirectedGraph(E=edges)    # Parameter V is optional
print(ug.V)                      # Output: {'Sagunt', 'València', 'Castelló'}
print(ug.succs('Sagunt'))        # Output: {'València', 'Castelló'}



from abc import abstractmethod, ABC


class Axis:
    X = 0
    Y = 1


class KDTree(ABC):
    @abstractmethod
    def pretty(self, level: int = 0) -> str: pass


class KDNode(KDTree):
    def __init__(self, axis: Axis, split_value: float, child1: KDTree, child2: KDTree):
        self.axis = axis  # Eje utilizado para separar sus hijos (Axis.X o Axis.Y)
        self.split_value = split_value  # Coordenada x o y (depende de axis)
        self.child1 = child1  # Hijo izquierdo o superior (Coordenada x o y (depende de axis) < split_value)
        self.child2 = child2  # Hijo dererecho o inferior (Coordenada x o y (depende de axis) >= split_value)

    def pretty(self, level: int = 0) -> str:
        return "       " * level + f"KDNode({self.axis}, {self.split_value},\n" + \
               self.child1.pretty(level + 1) + ",\n" + self.child2.pretty(level + 1) + "\n" + \
               "       " * level + ")"


class KDLeaf(KDTree):
    def __init__(self, point: tuple[float, float]):
        self.point = point

    def pretty(self, level: int = 0) -> str:
        return "       " * level + "KDLeaf({0})".format(self.point)



import sys
from abc import ABC, abstractmethod
from collections.abc import Callable, Iterable
from typing import Generic, TypeVar, Optional, Union

TVertex = TypeVar("TVertex")        # Generic Vertex
TEdge = tuple[TVertex, TVertex]     # Generic Edge
Weight = Union[int, float]

# UndirectedGraph, Digraph -----------------------------------------------------


# Abstract class (no instances)
class IGraph(ABC, Generic[TVertex]):
    @abstractmethod
    def is_directed(self) -> bool:
        pass

    def __init__(self, V: Iterable[TVertex] = None, E: Iterable[TEdge] = ()):
        self._e: list[TEdge] = list(E)
        if V is None:
            self._v: set[TVertex] = set(v for e in self._e for v in e)
        else:
            self._v: set[TVertex] = set(V)

        self._s: dict[TVertex, set[TVertex]] = dict((v, set()) for v in self._v)
        self._p: dict[TVertex, set[TVertex]] = dict((v, set()) for v in self._v) if self.is_directed() else self._s

        to_delete = []
        # Crea los diccionarioos self._s y self._p
        for (u, v) in self._e:
            if u == v:
                sys.stderr.write(f"WARNING: Discarted edge from a vertex to itself {(u, u)}\n")
                continue

            if u not in self._v:
                raise TypeError(f"Vertex {u} from edge {(u, v)} not in vertex set")
            if v not in self._v:
                raise TypeError(f"Vertex {v} from edge {(u, v)} not in vertex set")

            if v in self._s[u]:
                sys.stderr.write(f"WARNING: Discarted repeated edge from {u} to {v}\n")
                to_delete.append((u, v))
            else:
                self._s[u].add(v)
                self._p[v].add(u)

        for e in to_delete:
            self._e.remove(e)

    def succs(self, v: TVertex) -> set[TVertex]:
        return self._s[v]

    def preds(self, v: TVertex) -> set[TVertex]:
        return self._p[v]

    def out_degree(self, u: TVertex) -> int:
        return len(self._s[u])

    def in_degree(self, v: TVertex) -> int:
        return len(self._p[v])

    @property
    def V(self) -> set[TVertex]:
        return self._v

    @property
    def E(self) -> list[TEdge]:
        return self._e

    def add_vertex(self, v: TVertex):
        if v in self._v: return
        if v in self._s: raise Exception('Impossible 1')
        if v in self._p: raise Exception('Impossible 2')
        self._v.add(v)
        self._s[v] = set()
        self._p[v] = set()

    def remove_vertex(self, v: TVertex):
        if v not in self._v: return
        if v not in self._s: raise Exception('Impossible 3')
        if v not in self._p: raise Exception('Impossible 4')
        self._v.remove(v)

        for suc in self._s[v]:
            self._p[suc].remove(v)
        del self._s[v]

        e = [(u, w) for (u, w) in self._e if u != v and w != v]
        self._e = e

    def add_edge(self, e: TEdge):
        u, v = e
        self.add_vertex(u)
        self.add_vertex(v)

        self._e.append(e)

        self._s[u].add(v)
        self._p[v].add(u)

    def remove_edge(self, e: TEdge):
        u, v = e
        if u in self._s: self._s[u].discard(v)
        if v in self._p: self._p[v].discard(u)
        if e in self._e: self._e.remove(e)
        if not self.is_directed() and (v, u) in self._e:
            self._e.remove((v, u))

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(V={list(self.V)}, E={list(self.E)})"


class UndirectedGraph(IGraph[TVertex]):
    def is_directed(self) -> bool:
        return False


class Digraph(IGraph[TVertex]):
    def is_directed(self) -> bool:
        return True


# WeightingFunction -----------------------------------------------------------------


class WeightingFunction(Generic[TVertex], dict[TEdge, Weight], Callable[[TVertex, TVertex], Weight]):
    def __init__(self, data: Union[Iterable[tuple[TEdge, Weight]], dict[TEdge, Weight]], symmetrical: bool = False):
        super().__init__(data)
        self.symmetrical = symmetrical
        to_delete = []
        for (u, v) in self.keys():
            if u == v:
                sys.stderr.write(
                    f"WARNING - WeightingFunction: Edge from a vertex to itself {(u, u)} with weight {self[u, u]}.\n")
            elif symmetrical and (v, u) in self:
                if self[u, v] != self[v, u]:
                    raise ValueError("{!r} is different from {!r}".format((u, v), (v, u)))
                sys.stderr.write(
                    f"WARNING - WeightingFunction: simetrical with repeated edge: {(u, v)} and {(v, u)}. Discarted {(v, u)}.\n")
                to_delete.append((v, u))  # dejar solo una arista en grafos no dirigidos
        for (v, u) in to_delete:
            del self[v, u]

    def __call__(self, u: Union[TVertex, TEdge], v: Optional[TVertex] = None) -> Weight:
        if v is None:
            u, v = u
        # if u == v:
        #    return 0
        if (u, v) in self:
            return self[u, v]
        elif self.symmetrical:
            if (v, u) in self:
                return self[v, u]
            raise KeyError(repr((u, v)) + " nor " + repr((v, u)))
        raise KeyError(repr((u, v)))


if __name__ == '__main__':
    g = Digraph(E=[(1, 2)])
    print(g)
    print(isinstance(g, IGraph))
    print(isinstance(g, Digraph))
    print(isinstance(g, UndirectedGraph))



# coding: latin1
import unittest
from random import seed, shuffle

from algoritmia.datastructures.priorityqueues import MinHeap, MaxHeap


class TestHeap(unittest.TestCase):
    def setUp(self):
        self.minh1 = MinHeap()
        self.maxh1 = MaxHeap()
        a = list(range(10)) * 2
        seed(0)
        shuffle(a)
        self.minh2 = MinHeap(a)
        self.maxh2 = MaxHeap(a)

    def test_ctor(self):
        a = MinHeap([1, 2, 5, 4, 0, 3, 9, 7, 6, 8])
        b = MaxHeap([1, 2, 5, 4, 0, 3, 9, 7, 6, 8])
        for i in range(10):
            v = a.extract_opt()
            self.assertEqual(v, i)
            v = b.extract_opt()
            self.assertEqual(v, 9 - i)
        a = MinHeap()
        b = MaxHeap()
        self.assertRaises(IndexError, a.extract_opt)
        self.assertRaises(IndexError, b.extract_opt)

    def test_add(self):
        self.minh1.add(-1)
        self.assertEqual(self.minh1.opt(), -1)
        self.minh2.add(-1)
        self.assertEqual(self.minh2.opt(), -1)
        self.maxh1.add(100)
        self.assertEqual(self.maxh1.opt(), 100)
        self.maxh2.add(100)
        self.assertEqual(self.maxh2.opt(), 100)

    def test_extract_opt(self):
        for i in range(10):
            self.assertEqual(i, self.minh2.extract_opt())
            self.assertEqual(i, self.minh2.extract_opt())
        for i in range(9, -1, -1):
            self.assertEqual(i, self.maxh2.extract_opt())
            self.assertEqual(i, self.maxh2.extract_opt())
        self.assertEqual(len(self.minh2), 0)
        self.assertEqual(len(self.maxh2), 0)

    def test_MinHeap(self):
        mh = MinHeap(list(self.minh2))
        self.assertEqual(list(self.minh2), list(mh))

    def test_MaxHeap(self):
        mh = MaxHeap(list(self.maxh2))
        self.assertEqual(list(self.maxh2), list(mh))

    def test_iter(self):
        self.assertEqual(list(sorted(self.minh2)), list(sorted(list(range(10)) * 2)))

    def test_repr(self):
        self.assertEqual(list(sorted(self.minh2)), list(sorted(eval(repr(self.minh2)))))


if __name__ == "__main__":
    # import sys;sys.argv = ['', 'Test.testPoint2D']
    unittest.main()



# coding: latin1
import unittest
from random import seed, shuffle, randrange

from algoritmia.datastructures.doubleendedpriorityqueues import MinMaxIntervalHeap


class TestIntervalHeap(unittest.TestCase):
    def setUp(self):
        a = list(range(10)) * 2
        seed(0)
        shuffle(a)
        self.ih = MinMaxIntervalHeap(a)

    def test_ctor_fromIterable_shouldGiveElementsInOrder(self):
        a = MinMaxIntervalHeap([1, 2, 5, 4, 0, 3, 9, 7, 6, 8])
        for i in range(5):
            v = a.extract_min()
            self.assertEqual(v, i)
            v = a.extract_max()
            self.assertEqual(v, 9 - i)
        self.assertRaises(IndexError, a.extract_min)
        self.assertRaises(IndexError, a.extract_max)

    def test_iter_onIntervalHeap_shouldGiveAllItsElements(self):
        self.assertEqual(list(sorted(self.ih)), list(sorted(list(range(10)) * 2)))

    def test_stress(self):
        seed(0)
        d = []
        ih1 = MinMaxIntervalHeap()
        ih2 = MinMaxIntervalHeap()
        for i in range(200):
            v = randrange(1000)
            d.append(v)
            ih1.add(v)
            ih2.add(v)
        self.assertEqual(len(ih1), 200)
        d.sort()
        self.assertEqual(d, list(sorted(ih1)))
        for i in range(len(d)):
            self.assertEqual(ih1.min(), d[i])
            self.assertEqual(ih2.max(), d[-i - 1])
            v = ih1.extract_min()
            w = ih2.extract_max()
            self.assertEqual(v, d[i])
            self.assertEqual(w, d[-i - 1])

    def test_repr_onIntervalHeap_shouldReturnEvaluableExpression(self):
        self.assertEqual(list(sorted(self.ih)), list(sorted(eval(repr(self.ih)))))


if __name__ == "__main__":
    # import sys;sys.argv = ['', 'Test.testName']
    unittest.main()



import unittest

from algoritmia.datastructures.doubleendedprioritymaps import MinMaxIntervalHeapMap


class TestMinMaxPriorityDicts(unittest.TestCase):
    def setUp(self):
        self.mpd1 = MinMaxIntervalHeapMap()
        self.pairs = [('a', 1), ('z', 10), ('b', 5), ('c', 8), ('d', 12), ('e', 10)]
        self.mpd2 = MinMaxIntervalHeapMap(self.pairs)
        self.pairdict = dict(self.pairs)
        self.mpd3 = MinMaxIntervalHeapMap(self.pairdict)

    def test_ctor(self):
        self.assertEqual(len(self.mpd1), 0)
        self.assertEqual(len(self.mpd2), 6)
        self.assertEqual(len(self.mpd2), 6)

    def test_opt(self):
        self.assertRaises(IndexError, self.mpd1.opt)
        for i in (1, 5, 8, 10, 10, 12):
            self.assertEqual(self.mpd2[self.mpd2.opt()], i)
            self.mpd2.extract_opt()
            self.assertEqual(self.mpd3[self.mpd3.opt()], i)
            self.mpd3.extract_opt()

    def test_getitem(self):
        for k, v in self.pairs:
            self.assertEqual(self.mpd2[k], v)
            self.assertEqual(self.mpd3[k], v)
        self.assertRaises(KeyError, self.mpd2.__getitem__, 'xx')
        self.assertRaises(KeyError, self.mpd3.__getitem__, 'xx')

    def test_setitem(self):
        for pd in self.mpd1, self.mpd2, self.mpd3:
            for k, v in pd.items():
                pd[k] = pd[k] + 1
                self.assertEqual(pd[k], v + 1)
                pd[k] = pd[k] - 10
                self.assertEqual(pd[k], v - 9)
            for k, v in ('xx', 10), ('yy', 100):
                pd[k] = v
                self.assertEqual(pd[k], v)

    def test_opt_value_and_item(self):
        for pd in self.mpd1, self.mpd2, self.mpd3:
            if len(pd) > 0:
                self.assertEqual(pd.opt_item()[1], pd.opt_value())
                self.assertEqual(pd.opt_item()[0], pd.opt())
            else:
                self.assertRaises(IndexError, pd.opt_item)

    def test_contains(self):
        for pd in self.mpd2, self.mpd3:
            for k, v in self.pairs:
                self.assertTrue(k in pd)
        for pd in [self.mpd1]:
            for k, v in self.pairs:
                self.assertFalse(k in pd)

    def test_del(self):
        for pd in self.mpd2, self.mpd3:
            for k, v in self.pairs:
                del pd[k]
            self.assertEqual(len(pd), 0)

    def test_repr(self):
        i = 0
        for pd in self.mpd1, self.mpd2, self.mpd3:
            i += 1
            if len(pd) > 0:
                self.assertEqual(list(sorted(self.pairs)), list(sorted(eval(repr(pd)).items())))


class TestMinMaxPriorityDicts2(unittest.TestCase):
    def setUp(self):
        self.pairs = [('a', 1), ('z', 11), ('b', 5), ('c', 8), ('d', 12), ('e', 10)]
        self.a = MinMaxIntervalHeapMap(self.pairs)

    def test_len(self):
        self.assertEqual(len(self.pairs), len(self.a))

    def test_opt_worst(self):
        self.assertEqual('a', self.a.opt())
        self.assertEqual(1, self.a.opt_value())
        self.assertEqual(('a', 1), self.a.opt_item())
        self.assertEqual('d', self.a.worst())
        self.assertEqual(12, self.a.worst_value())
        self.assertEqual(('d', 12), self.a.worst_item())

        self.assertEqual('a', self.a.extract_min())
        self.assertEqual(5, self.a.extract_min_value())
        self.assertEqual(('c', 8), self.a.extract_min_item())
        self.assertEqual('d', self.a.extract_max())
        self.assertEqual(11, self.a.extract_max_value())
        self.assertEqual(('e', 10), self.a.extract_max_item())


if __name__ == "__main__":
    # import sys;sys.argv = ['', 'Test.testName']
    unittest.main()



from abc import abstractmethod, ABC
from collections.abc import Iterable, Sized
from typing import TypeVar, Generic

from algoritmia.datastructures.linkedlists import LinkedList

T = TypeVar('T')


# Fifo() utiliza una lista LinkedList para tener coste O(1) en el pop()
# Lifo() utiliza una lista normal de Python


class IQueue(ABC, Sized, Generic[T]):
    def __init__(self, mylist: list):
        self._list = mylist

    @abstractmethod
    def push(self, item: T): pass

    @abstractmethod
    def pop(self) -> T: pass

    @abstractmethod
    def top(self) -> T: pass

    def __len__(self) -> int:
        return len(self._list)

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({list(self._list)})'


class Fifo(IQueue):
    def __init__(self, data: Iterable[T] = ()):
        super().__init__(LinkedList(data))

    def push(self, item: T):
        self._list.append(item)

    def pop(self) -> T:
        v = self._list[0]
        del self._list[0]
        return v

    def top(self) -> T:
        return self._list[0]


class Lifo(IQueue):
    def __init__(self, data: Iterable[T] = ()):
        super().__init__(list(data))

    def push(self, item: T):
        self._list.append(item)

    def pop(self) -> T:
        return self._list.pop()

    def top(self) -> T:
        return self._list[-1]


if __name__ == '__main__':
    for q in [Fifo(), Lifo()]:
        for i in range(3):
            q.push(i)
        print(q)
        for i in range(3):
            print(q.pop())



import random

from algoritmia.datastructures.graphs import UndirectedGraph
from algoritmia.datastructures.mergefindsets import MergeFindSet
from algoritmia.viewers.labyrinth_viewer import LabyrinthViewer

Vertex = tuple[int, int]
Edge = tuple[Vertex, Vertex]


def create_labyrinth_mfset(num_rows: int, num_cols: int, extra_corridors: int = 0) -> UndirectedGraph[Vertex]:
    # Crear lista de vértices
    v = [(r, c) for r in range(num_rows) for c in range(num_cols)]

    # Crear lista de aristas entre vértices vecinos
    w = []
    for (r, c) in v:
        if r + 1 < num_rows: w.append(((r, c), (r + 1, c)))
        if c + 1 < num_cols: w.append(((r, c), (r, c + 1)))

    # Barajar la lista de aristas entre vértices vecinos
    random.shuffle(w)
    e = []

    # Crear un MFSet vacío
    mfs = MergeFindSet()

    # Añadir los elementos de 'v' al MFSet
    for vertice in v: mfs.add(vertice)

    # Recorrer aristas en 'w',
    # obtener 'find' de los vertices de cada arista (v1,v2),
    # si 'v1' y 'v2' están en clases diferentes: hacer 'merge' y añadir arista a 'e'
    for (v1, v2) in w:
        if mfs.find(v1) != mfs.find(v2):
            mfs.merge(v1, v2)
            e.append((v1, v2))
        elif extra_corridors > 0:
            e.append((v1, v2))
            extra_corridors -= 1

    return UndirectedGraph(E=e)


if __name__ == '__main__':
    random.seed(42)

    cell_size = 60
    num_rows, num_cols = 6, 8

    g = create_labyrinth_mfset(num_rows, num_cols, extra_corridors=0) # 3

    lv = LabyrinthViewer(g, margin=10,
                         canvas_width=num_cols * cell_size + 20,
                         canvas_height=num_rows * cell_size + 20)
    lv.run()



from collections.abc import Iterator

from algoritmia.algorithms.traversers import TVertexTraverser
from algoritmia.datastructures.graphs import UndirectedGraph, TVertex

TCC = list[TVertex]  # Generic Connected Component


def connected_components(g: UndirectedGraph[TVertex],
                         vertex_traverser: TVertexTraverser) -> Iterator[TCC]:
    pending_vertices = set(g.V)
    while len(pending_vertices) > 0:
        u = pending_vertices.pop()
        visited_vertices = list(vertex_traverser(g, u))
        pending_vertices -= set(visited_vertices)
        yield visited_vertices


if __name__ == '__main__':
    from traversers import bf_vertex_traverser, df_vertex_traverser
    Vertex = tuple[int, int]
    Edge = tuple[Vertex, Vertex]

    edges: list[Edge] = [((0, 0), (0, 1)), ((0, 2), (0, 3)), ((1, 0), (1, 1)),
                         ((2, 2), (2, 3)), ((0, 1), (1, 1)), ((0, 2), (1, 2)),
                         ((1, 2), (2, 2)), ((2, 0), (2, 1)), ((0, 3), (1, 3))]

    my_graph = UndirectedGraph(E=edges)

    ccs_bf = list(connected_components(my_graph, bf_vertex_traverser))
    print("Breath first::", ccs_bf)

    ccs_df = list(connected_components(my_graph, df_vertex_traverser))
    print("Depth first:", ccs_df)



from __future__ import annotations

from collections.abc import Iterable, Iterator
from dataclasses import dataclass

from algoritmia.datastructures.graphs import UndirectedGraph, TVertex, WeightingFunction
from algoritmia.schemes.bt_scheme import DecisionSequence, bt_solve
from algoritmia.schemes.bt_scheme import ScoredDecisionSequence, bt_min_solve

Decision = TVertex
Score = float

# 'bt_solve' y 'bt_vc_solve' devuelven un Iterator del tipo devuelto por el método 'solution' de
# la clase 'DecisionSequence', cuya implementación por defecto devuelve una tupla con las decisiones:
SolutionDS = tuple[Decision, ...]

# 'bt_min_solve' y 'bt_max_solve' devuelven un Iterator del tipo devuelto por el método 'solution' de
# la clase 'ScoredDecisionSequence', cuya implementación por defecto devuelve una tupla de dos elementos:
# el Score y una tupla con las decisones:
SolutionSDS = tuple[Score, tuple[Decision, ...]]


def hamiltoniancycle_solve(graph: UndirectedGraph[TVertex]) -> Iterator[SolutionDS]:
    class HamiltonianCycleDS(DecisionSequence[TVertex]):
        def is_solution(self) -> bool:
            ds = self.decisions()
            return len(ds) == len(graph.V) and ds[0] in graph.succs(ds[-1])

        def successors(self) -> Iterable[HamiltonianCycleDS]:
            ds = self.decisions()
            if len(ds) < len(graph.V):
                for v in graph.succs(ds[-1]):
                    if v not in ds:
                        yield self.add_decision(v)

    v_initial = next(iter(graph.V))  # Para coger un elemento de un conjunto sin sacarlo
    initial_ds = HamiltonianCycleDS().add_decision(v_initial)
    return bt_solve(initial_ds)


def hamiltoniancycle_opt_solve(graph: UndirectedGraph[TVertex],
                               wf: WeightingFunction[TVertex]) -> Iterator[SolutionSDS]:
    @dataclass
    class Extra:
        ls: int = 0

    class HamiltonianCycleDS(ScoredDecisionSequence[TVertex]):
        def is_solution(self) -> bool:
            ds = self.decisions()
            return len(ds) == len(graph.V) and ds[0] in graph.succs(ds[-1])

        def successors(self) -> Iterable[HamiltonianCycleDS]:
            ds = self.decisions()
            if len(ds) < len(graph.V):
                for v in graph.succs(ds[-1]):
                    if v not in ds:
                        new_extra = Extra(self.extra.ls + wf(ds[-1], v))
                        yield self.add_decision(v, new_extra)

        # Sobreescribimos 'state()'
        def state(self) -> tuple[TVertex, tuple[Vertex, ...]]:
            ds = list(self.decisions())
            return ds[-1], tuple(sorted(ds))

        def score(self) -> float:
            if len(self) < len(graph.V): return self.extra.ls
            ds = self.decisions()
            if ds[0] in graph.succs(ds[-1]):
                return self.extra.ls + wf(ds[0], ds[-1])
            return float('infinity')

    v_initial = next(iter(graph.V))  # Para coger un elemento de un conjunto sin sacarlo
    initial_ds = HamiltonianCycleDS(Extra()).add_decision(v_initial, Extra())
    return bt_min_solve(initial_ds)


# Programa principal ---------------------------
if __name__ == "__main__":
    Vertex = int
    Edge = tuple[Vertex, Vertex]
    edges: list[Edge] = [(0, 2), (0, 3), (0, 9), (1, 3), (1, 4), (1, 8), (2, 3), (2, 5), (3, 4),
                         (3, 6), (4, 7), (5, 6), (5, 8), (6, 7), (6, 8), (6, 9)]
    g = UndirectedGraph(E=edges)

    print("hamiltoniancycle_solve:")
    for solution in hamiltoniancycle_solve(g):
        print(solution)
    print()

    print("hamiltoniancycle_opt_solve:")
    d: dict[tuple[int, int], int] = {}
    for (u2, v2) in g.E:
        d[u2, v2] = abs(u2 - v2)
    wf2 = WeightingFunction((e for e in d.items()), symmetrical=True)
    for solution_opt in hamiltoniancycle_opt_solve(g, wf2):
        print(solution_opt)



from abc import abstractmethod, ABC
from collections.abc import Callable, Iterable, Iterator, Sequence
from itertools import chain, repeat
from math import log, sqrt
from typing import Generic, TypeVar, Union, Optional

K = TypeVar('K')
T = TypeVar('T')


class IPriorityMap(ABC, dict[K, T]):
    @abstractmethod
    def opt(self) -> K: pass

    @abstractmethod
    def opt_item(self) -> tuple[K, T]: pass

    @abstractmethod
    def opt_value(self) -> T: pass

    @abstractmethod
    def extract_opt(self) -> K: pass

    @abstractmethod
    def extract_opt_item(self) -> tuple[K, T]: pass


# -------------------------------------------------------------------------

class MinHeapMap(IPriorityMap[K, T]):
    _opt: Callable[[T, T], T] = min

    def __init__(self, data: Union[Iterable[tuple[K, T]], dict[K, T]] = (), capacity: int = 0):
        super().__init__()
        # self._opt = opt
        self._index = {}
        if isinstance(data, dict):
            data = data.items()
        elif not isinstance(data, Sequence):
            data = tuple(data)
        for (i, (key, _)) in enumerate(data):
            self._index[key] = i + 1
        self._size = len(data)
        self._heap: list[Optional[tuple[T, K]]] = list(chain((None,), ((v, k) for (k, v) in data),
                                                             repeat(None, max(0, capacity - self._size))))
        for i in range(self._size // 2, 0, -1):
            self._heapify(i)

    def _heapify(self, i: int):
        while True:
            left: int = 2 * i
            right: int = 2 * i + 1
            if left <= self._size and self._opt(self._heap[left], self._heap[i]) != self._heap[i]:
                best = left
            else:
                best = i
            if right <= self._size and self._opt(self._heap[right], self._heap[best]) != self._heap[best]:
                best = right
            if best == i:
                break
            self._index[self._heap[i][1]], self._index[self._heap[best][1]] = best, i
            self._heap[i], self._heap[best] = self._heap[best], self._heap[i]
            i = best

    def _bubble_up(self, i: int):
        p = i // 2
        while i > 1 and self._opt(self._heap[i], self._heap[p]) != self._heap[p]:
            self._index[self._heap[i][1]], self._index[self._heap[p][1]] = p, i
            self._heap[i], self._heap[p] = self._heap[p], self._heap[i]
            i, p = p, p // 2

    def opt(self) -> K:
        if self._size == 0:
            raise IndexError('opt from an empty priority dict')
        return self._heap[1][1]

    def opt_item(self) -> tuple[K, T]:
        if self._size == 0:
            raise IndexError('opt from an empty priority dict')
        return self._heap[1][1], self._heap[1][0]

    def opt_value(self) -> T:
        if self._size == 0:
            raise IndexError('opt from an empty priority dict')
        return self._heap[1][0]

    def extract_opt(self) -> K:
        return self.extract_opt_item()[0]

    def extract_opt_item(self) -> tuple[K, T]:
        m = self.opt_item()
        if self._size > 1:
            self._heap[1], self._index[self._heap[self._size][1]] = self._heap[self._size], 1
            self._heap[self._size] = None
        self._size -= 1
        if self._size > 1:
            self._heapify(1)
        del self._index[m[0]]
        return m

    def __contains__(self, key: K) -> bool:
        return key in self._index

    def __getitem__(self, key: K) -> T:
        return self._heap[self._index[key]][0]

    def __setitem__(self, key: K, score: T) -> T:
        if key in self._index:
            i = self._index[key]
            if score == self._heap[i][0]:
                return score
            if self._opt(score, self._heap[i][0]) != score:
                self._heap[i] = (score, key)
                self._heapify(i)
                return score
        else:
            if self._size + 1 >= len(self._heap):
                self._heap.append(None)
            self._index[key] = i = self._size = self._size + 1
        self._heap[i] = (score, key)
        self._bubble_up(i)
        return score

    def __delitem__(self, key: K):
        if key not in self._index:
            raise KeyError(key)
        i = self._index[key]
        self._heap[i], self._index[self._heap[self._size][1]] = self._heap[self._size], i
        self._size -= 1
        self._heapify(i)
        if i > 1 and self._heap[i] < self._heap[i // 2]:
            p = i // 2
            while i > 1 and self._heap[i] < self._heap[p]:
                self._index[self._heap[i][1]], self._index[self._heap[p][1]] = p, i
                self._heap[i], self._heap[p] = self._heap[p], self._heap[i]
                i, p = p, p // 2
        else:
            self._heapify(i)
        del self._index[key]

    def keys(self) -> Iterator[K]:
        for key in self._index:
            yield key

    def values(self) -> Iterator[T]:
        for key in self._index:
            yield self._heap[self._index[key]][0]

    def items(self) -> Iterator[tuple[K, T]]:
        for key in self._index:
            yield self._heap[self._index[key]][1], self._heap[self._index[key]][0]

    def get(self, key: K, default: T = None) -> T:
        if key in self._index:
            return self[key]
        return default

    def setdefault(self, key: K, default: T = None) -> T:
        if key in self._index:
            return self[key]
        self[key] = default
        return default

    def __iter__(self) -> Iterator[K]:
        for key in self._index:
            yield key

    def __len__(self) -> int:
        return self._size

    def __repr__(self) -> str:
        if self.__class__.__name__ == 'HeapMap':
            b = 'min' if self._opt == min else 'max'
            return 'HeapMap({}, {!r})'.format(b, [(k, self[k]) for k in self])
        return '{}({!r})'.format(self.__class__.__name__, [(k, self[k]) for k in self])


class MaxHeapMap(MinHeapMap):
    _opt = max


# -------------------------------------------------------------------------

class FibNode(Generic[K, T]):
    def __init__(self, key: K, value: T):
        self.parent = self.child = None
        self.left = self.right = self
        self.key = key
        self.value = value
        self.degree = 0
        self.mark = False


class MinFibonacciHeap(IPriorityMap[K, T]):
    _opt: Callable[[T, T], T] = min

    def __init__(self, data: Iterable[tuple[K, T]] = ()):
        super().__init__()
        self._size = 0
        self._minroot: Optional[FibNode[K, T]] = None
        self._map = dict()
        self._opt2 = lambda a, b: None if a is None or b is None else self._opt(a, b)
        for key, value in data:
            self.add(key, value)

    def __getitem__(self, key: K) -> T:
        return self._map[key].value

    def opt_item(self) -> tuple[K, T]:
        return self._minroot.key, self._minroot.value

    def opt(self) -> K:
        return self._minroot.key

    def opt_value(self) -> T:
        return self._minroot.value

    def add(self, key: K, value: T):
        node = self._map[key] = FibNode(key, value)
        if self._minroot is None:
            self._minroot = node
        else:
            node.left, node.right = self._minroot, self._minroot.right
            self._minroot.right = node.right.left = node
            if self._opt2(node.value, self._minroot.value) != self._minroot.value:
                self._minroot = node
        self._size += 1

    def extract_opt(self) -> K:
        opt = self.opt()
        self._remove_opt()
        return opt

    def extract_opt_item(self) -> tuple[K, T]:
        item = (self._minroot.key, self._minroot.value)
        self._remove_opt()
        return item

    def _remove_opt(self):
        z = self._minroot
        del self._map[z.key]
        if z is not None:
            nchildren = z.degree
            x = z.child
            while nchildren > 0:
                t = x.right
                x.left.right, x.right.left = x.right, x.left
                x.left, x.right = self._minroot, self._minroot.right
                self._minroot.right = x.right.left = x
                x.parent = None
                x = t
                nchildren -= 1
            z.left.right, z.right.left = z.right, z.left
            if z == z.right:
                self._minroot = None
            else:
                self._minroot = z.right
                self._consolidate()
            self._size -= 1

    _philog = log((1 + sqrt(5)) / 2)

    def _consolidate(self):
        a: list[Optional[FibNode[K, T]]] = [None] * int(log(self._size) / self._philog)
        n_roots = self._count_roots()
        x = self._minroot
        while n_roots > 0:
            d = x.degree
            next0 = x.right
            while a[d] is not None:
                y = a[d]
                if self._opt2(x.value, y.value) != x.value: x, y = y, x
                self._link(y, x)
                a[d] = None
                d += 1
            a[d] = x
            x = next0
            n_roots -= 1

        self._minroot = None
        for x in a:
            if x is not None:
                if self._minroot is not None:
                    x.left.right, x.right.left = x.right, x.left
                    x.left, x.right = self._minroot, self._minroot.right
                    self._minroot.right = x
                    x.right.left = x
                    if self._opt2(x.value, self._minroot.value) != self._minroot.value:
                        self._minroot = x
                else:
                    self._minroot = x

    def _count_roots(self) -> int:
        n_roots = 0
        x = self._minroot
        if x is not None:
            n_roots += 1
            x = x.right
            while x != self._minroot:
                n_roots += 1
                x = x.right
        return n_roots

    @staticmethod
    def _link(y: FibNode[K, T], x: FibNode[K, T]):
        y.left.right, y.right.left = y.right, y.left
        y.parent = x
        if x.child is None:
            x.child = y.right = y.left = y
        else:
            y.left, y.right = x.child, x.child.right
            x.child.right = y.right.left = y
        x.degree += 1
        y.mark = False

    def __setitem__(self, key: K, value: T):
        if key not in self._map:
            self.add(key, value)
        else:
            self._improve_value(key, value)

    def _improve_value(self, key: K, value: T):
        node = self._map[key]
        if self._opt2(value, node.value) != value:
            raise ValueError("{} at {} does not improve {}".format(value, key, node.value))
        node.value = value
        parent = node.parent
        if parent is not None and self._opt2(node.value, parent.value) != parent.value:
            self._cut(node, parent)
            self._cascading_cut(parent)
        if self._opt2(node.value, self._minroot.value) != self._minroot.value:
            self._minroot = node

    def _cut(self, x: FibNode[K, T], y: FibNode[K, T]):
        x.left.right, x.right.left = x.right, x.left
        y.degree -= 1
        if y.child == x: y.child = x.right
        if y.degree == 0: y.child = None
        x.left, x.right = self._minroot, self._minroot.right
        self._minroot.right = x
        x.right.left = x
        x.parent = None
        x.mark = False

    def _cascading_cut(self, node: FibNode[K, T]):
        parent = node.parent
        if parent is not None:
            if not node.mark:
                node.mark = True
            else:
                self._cut(node, parent)
                self._cascading_cut(parent)

    def __delitem__(self, key: K):
        self._improve_value(key, None)
        self._remove_opt()

    def keys(self) -> Iterable[K]:
        return self._map.keys()

    def values(self) -> Iterator[T]:
        for key in self._map: yield self._map[key].value

    def items(self) -> Iterator[tuple[K, T]]:
        for key in self._map:
            yield key, self._map[key].value

    def get(self, key: K, default: T = None) -> T:
        if key in self._map: return self[key]
        return default

    def setdefault(self, key: K, default: T = None):
        if key in self._map: return self[key]
        self[key] = default
        return default

    def __contains__(self, key: K) -> bool:
        return key in self._map

    def __len__(self) -> "int":
        return self._size

    def __iter__(self) -> Iterator[K]:
        for key in self._map: yield key

    def __repr__(self) -> str:
        if self.__class__.__name__ == 'FibonacciHeap':
            b = 'min' if self._opt2 == min else 'max'
            return 'FibonacciHeap({}, {!r})'.format(b, [(k, self[k]) for k in list(self)])
        return '{}({!r})'.format(self.__class__.__name__, [(k, self[k]) for k in list(self)])


class MaxFibonacciHeap(MinFibonacciHeap):
    _opt = max

# -------------------------------------------------------------------------



from abc import abstractmethod
from collections.abc import Iterator, Collection
from typing import TypeVar

from algoritmia.datastructures.priorityqueues import IPriorityQueue

T = TypeVar('T')


class IDoubleEndedPriorityQueue(IPriorityQueue[T]):
    @abstractmethod
    def worst(self) -> T: pass

    @abstractmethod
    def extract_worst(self) -> T: pass


class IntervalHeap(IDoubleEndedPriorityQueue):
    def __init__(self, data: Collection[T] = (), capacity: int = 0):
        capacity = max(capacity, len(data))
        self._heap = list(data) + [None] * (capacity - len(data))
        self._size = len(data)
        for v in range(0, self._size, 2):
            if v + 1 < self._size: self._swap(v)
        last_parent = self._parent(self._size - 1)
        for v in range(last_parent, -1, -2):
            self._heapify_min(v)
            self._heapify_max(v)

    def _swap(self, i: int) -> bool:
        if self._heap[i] > self._heap[i + 1]:
            self._heap[i], self._heap[i + 1] = self._heap[i + 1], self._heap[i]
            return True
        return False

    @staticmethod
    def _parent(i: int) -> int:
        return ((i // 2 - 1) // 2) * 2

    def _heapify_min(self, i: int):
        smallest = i
        while True:
            for j in self._children(i):
                if self._heap[j] < self._heap[smallest]: smallest = j
            if smallest == i: break
            self._heap[smallest], self._heap[i] = self._heap[i], self._heap[smallest]
            i = smallest
            if i + 1 < self._size:
                self._swap(i)

    def _heapify_max(self, i: int):
        largest = i + 1
        while True:
            for j in self._children(i):
                if j + 1 < self._size:
                    if self._heap[j + 1] > self._heap[largest]: largest = j + 1
                else:
                    if self._heap[j] > self._heap[largest]: largest = j
            if largest == i + 1: break
            self._heap[largest], self._heap[i + 1] = self._heap[i + 1], self._heap[largest]
            i = largest // 2 * 2
            if i + 1 < self._size:
                self._swap(i)

    def _children(self, i: int) -> Iterator[int]:
        j = 2 * (i + 1)
        if j < self._size: yield j
        j += 2
        if j < self._size: yield j

    def add(self, v: T):
        if self._size == len(self._heap):
            self._heap.append(None)
        if self._size == 0:
            self._heap[0] = v
            self._size += 1
        else:
            self._heap[self._size] = v
            self._size += 1
            if self._size % 2 == 0:
                i = self._size - 2
                if self._swap(i):
                    self._bubble_up_min(i)
                else:
                    self._bubble_up_max(i)
            else:
                i = self._size - 1
                parent = self._parent(i)
                if parent >= 0:
                    if self._heap[parent] <= v <= self._heap[parent + 1]: return
                    if v > self._heap[parent + 1]:
                        self._heap[parent + 1], self._heap[i] = self._heap[i], self._heap[parent + 1]
                        self._bubble_up_max(parent)
                    elif v < self._heap[parent]:
                        self._heap[parent], self._heap[i] = self._heap[i], self._heap[parent]
                        self._bubble_up_min(parent)

    def _bubble_up_max(self, i: int):
        while True:
            parent = self._parent(i)
            if parent < 0: return
            if self._heap[parent + 1] >= self._heap[i + 1]: return
            self._heap[parent + 1], self._heap[i + 1] = self._heap[i + 1], self._heap[parent + 1]
            i = parent

    def _bubble_up_min(self, i: int):
        while True:
            parent = self._parent(i)
            if parent < 0: return
            if self._heap[parent] <= self._heap[i]: return
            self._heap[parent], self._heap[i] = self._heap[i], self._heap[parent]
            i = parent

    def min(self) -> T:
        if self._size == 0: raise IndexError("Empty Interval Heap")
        return self._heap[0]

    def max(self) -> T:
        if self._size == 0: raise IndexError("Empty Interval Heap")
        if self._size == 1: return self._heap[0]
        return self._heap[1]

    def extract_min(self) -> T:
        if self._size == 0: raise IndexError("Empty Interval Heap")
        retval = self._heap[0]
        if self._size <= 2:
            if self._size == 1:
                self._heap[0] = None
            else:
                self._heap[0], self._heap[1] = self._heap[1], None
            self._size -= 1
            return retval
        self._heap[0], self._heap[self._size - 1] = self._heap[self._size - 1], None
        self._size -= 1
        self._heapify_min(0)
        return retval

    def extract_max(self) -> T:
        if self._size == 0: raise IndexError("Empty Interval Heap")
        if self._size == 1:
            retval = self._heap[0]
            self._heap[0] = None
            self._size -= 1
            return retval
        retval = self._heap[1]
        if self._size == 2:
            self._heap[1] = None
            self._size -= 1
            return retval
        self._heap[1], self._heap[self._size - 1] = self._heap[self._size - 1], None
        self._size -= 1
        self._heapify_max(0)
        return retval

    def __len__(self) -> int:
        return self._size

    def __iter__(self) -> Iterator[T]:
        for i in range(self._size):
            yield self._heap[i]

    def __repr__(self) -> str:
        return '{}({!r})'.format(self.__class__.__name__, self._heap[:self._size])

    opt = min
    extract_opt = extract_min

    worst = max
    extract_worst = extract_max


class MinMaxIntervalHeap(IntervalHeap, IDoubleEndedPriorityQueue):
    opt = IntervalHeap.min
    extract_opt = IntervalHeap.extract_min

    worst = IntervalHeap.max
    extract_worst = IntervalHeap.extract_max


class MaxMinIntervalHeap(IntervalHeap, IDoubleEndedPriorityQueue):
    opt = IntervalHeap.max
    extract_opt = IntervalHeap.extract_max

    worst = IntervalHeap.min
    extract_worst = IntervalHeap.extract_min



from __future__ import annotations

from collections.abc import Iterable
from dataclasses import dataclass
from random import seed, randint
from typing import Optional

from algoritmia.schemes.bab_scheme import BoundedDecisionSequence, bab_max_solve

Decision = int  # 0 o 1
Score = int

# 'bab_max_solve' devuelve un Iterator del tipo devuelto por el método 'solution' de
# la clase 'BoundedDecisionSequence', cuya implementación por defecto devuelve una tupla de dos elementos:
# el Score (el valor del contenido de la mochila) y una tupla con las decisones:
Solution = tuple[Score, tuple[Decision, ...]]


# Podriamos sobreescribir  el método 'solution()' en KnapsackBabDS para que devolviera también el peso.
# En ese caso el tipo de 'Solution' sería:
# Solution = tuple[Score, int, tuple[Decision, ...]]  # (value, weight, decisions)

def knapsack_bab_solve(weights: list[int],
                       values: list[int],
                       capacity: int) -> Optional[Solution]:
    @dataclass
    class Extra:
        weight: int = 0
        value: int = 0

    class KnapsackBabDS(BoundedDecisionSequence[Decision]):
        # Coge TODOS los objetos pendientes
        def calculate_opt_bound(self) -> Score:
            return self.extra.value + sum(values[len(self):])

        # No coge NINGUNO de los objetos pendientes
        def calculate_pes_bound(self) -> Score:
            return self.extra.value

        def is_solution(self) -> bool:
            return len(self) == len(values)

        def successors(self) -> Iterable[KnapsackBabDS]:
            n = len(self)
            if n < len(values):
                if weights[n] <= capacity - self.extra.weight:
                    new_extra = Extra(self.extra.weight + weights[n], self.extra.value + values[n])
                    yield self.add_decision(1, new_extra)
                yield self.add_decision(0, self.extra)

        # Podríamos sobreescribir 'solution()' para devolver también el peso:
        # def solution(self) -> Solution:
        #    return self.extra.value, self.extra.weight, self.decisions()

    initial_ds = KnapsackBabDS(Extra())
    return bab_max_solve(initial_ds)


# Versión con mejores cotas
def knapsack_bab_solve2(weights: list[int],
                        values: list[int],
                        capacity: int) -> Optional[Solution]:
    @dataclass
    class Extra:
        weight: int = 0
        value: int = 0

    class KnapsackBabDS(BoundedDecisionSequence[Decision]):
        # IMPLEMENTAR: resolver mochila continua para los objetos que quedan (tema 3)
        def calculate_opt_bound(self) -> int:
            value = self.extra.value
            weight = self.extra.weight
            for i in range(len(self), len(weights)):
                f = min(1, (capacity - weight) / weights[i])
                weight += f * weights[i]
                value += f * values[i]
                if f < 1:
                    break
            return value

        # IMPLEMENTAR: modificación del optimista sin fraccionar
        def calculate_pes_bound(self) -> int:
            value = self.extra.value
            weight = self.extra.weight
            for i in range(len(self), len(weights)):
                if weight + weights[i] <= capacity:
                    weight += weights[i]
                    value += values[i]
            return value

        def is_solution(self) -> bool:
            return len(self) == len(values)

        def successors(self) -> Iterable[KnapsackBabDS]:
            n = len(self)
            if n < len(values):
                if weights[n] <= capacity - self.extra.weight:
                    new_extra = Extra(self.extra.weight + weights[n], self.extra.value + values[n])
                    yield self.add_decision(1, new_extra)
                yield self.add_decision(0, self.extra)

        # Podríamos sobreescribir 'solution()' para devolver también el peso:
        # def solution(self) -> Solution:
        #    return self.extra.value, self.extra.weight, self.decisions()

    initial_ps = KnapsackBabDS(Extra())
    return bab_max_solve(initial_ps)


def sorted_by_dec_ratio(w_old: list[int], v_old: list[int]) -> tuple[list[int], list[int]]:
    idxs: list[int] = sorted(range(len(w_old)), key=lambda i: -v_old[i] / w_old[i])
    w_new = [w_old[i] for i in idxs]
    v_new = [v_old[i] for i in idxs]
    return w_new, v_new


# Crea los problemas con los objetos ordenados de mayor a menor beneficio por kilo
def create_knapsack_problem(num_objects: int) -> tuple[list[int], list[int], int]:
    seed(5)
    w_new = [randint(10, 100) for _ in range(num_objects)]
    v_new = [w_new[i] * randint(1, 4) for i in range(num_objects)]
    capacity = int(sum(w_new) * 0.3)
    weights, values = sorted_by_dec_ratio(w_new, v_new)
    return weights, values, capacity


# Main program -------------------------------------------------------
if __name__ == "__main__":
    # Cada instancia es una tupla (pesos, valores, capacidad):

    # Solution: value = 354, decisions = (1, 0, 1, 0, 0, 0))
    i1 = create_knapsack_problem(6)

    # Solution: value = 1118, decisions = (1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0)
    i2 = create_knapsack_problem(20)

    # Solution: value = 1830, decisions = (1, 1, 1, 1, 1, 0, 1, 0, 1, 1, 1, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0)
    i3 = create_knapsack_problem(35)

    for W, V, C in [i1, i2, i3]:
        print(f"Instancia:\n  Pesos = {W}\n  Valores = {V}\n  Capacidad = {C}")
        print(f"Solution:\n  {knapsack_bab_solve(W, V, C)}\n")
        #print(f"Solution:\n  {knapsack_bab_solve2(W, V, C)}\n")



"""
Version:  4.1 (29-sep-2022)
          4.0 (23-oct-2021)

@author: David Llorens (dllorens@uji.es)
         (c) Universitat Jaume I 2022
@license: GPL3
"""
from __future__ import annotations

from abc import ABC, abstractmethod
from collections.abc import Iterator, Iterable
from typing import TypeVar, Generic, Any, Union

infinity = float("infinity")

TDecision = TypeVar('TDecision', contravariant=True)
Solution = Any


# Esquema para BT básico --------------------------------------------------------------------------

class DecisionSequence(ABC, Generic[TDecision]):
    __slots__ = ("_decisions", "_extra_fields")

    def __init__(self, extra_fields=None, decisions: tuple[TDecision, ...] = ()):
        self._decisions = decisions
        self._extra_fields = extra_fields

    @property
    def extra(self) -> Any:
        return self._extra_fields

    def __len__(self) -> int:  # len(objeto) devuelve el número de decisiones del objeto
        return len(self._decisions)

    def add_decision(self, d: TDecision, extra_fields=None) -> DecisionSequence:
        return self.__class__(extra_fields, self._decisions + (d,))

    def decisions(self) -> tuple[TDecision, ...]:
        return self._decisions

    def solution(self) -> Solution:
        return self._decisions

    @abstractmethod
    def is_solution(self) -> bool:
        pass

    @abstractmethod
    def successors(self) -> Iterable[DecisionSequence]:
        pass


def bt_solve(initial_ds: DecisionSequence[TDecision]) -> Iterator[Solution]:
    def bt(ds: DecisionSequence[TDecision]) -> Iterator[Solution]:
        if ds.is_solution():
            yield ds.solution()
        for new_ds in ds.successors():
            yield from bt(new_ds)

    return bt(initial_ds)


#  Esquema para BT con control de visitados --------------------------------------------------------

class StateDecisionSequence(DecisionSequence[TDecision]):
    @abstractmethod
    def successors(self) -> Iterable[StateDecisionSequence]:
        pass

    def state(self) -> Any:
        # The returned object must be of an inmutable type
        return self._decisions


def bt_vc_solve(initial_ds: StateDecisionSequence[TDecision]) -> Iterator[Solution]:
    def bt(ds: StateDecisionSequence[TDecision]) -> Iterator[Solution]:
        seen.add(ds.state())  # El método state debe devolver un objeto inmutable
        if ds.is_solution():
            yield ds.solution()
        for new_ds in ds.successors():
            state = new_ds.state()
            if state not in seen:
                yield from bt(new_ds)

    seen = set()
    return bt(initial_ds)


# Esquema para BT para optimización ----------------------------------------------------------------

class ScoredDecisionSequence(StateDecisionSequence[TDecision]):
    @abstractmethod
    def successors(self) -> Iterable[ScoredDecisionSequence]:
        pass

    @abstractmethod
    def score(self) -> Union[int, float]:
        # result of applying the objective function to the partial solution
        pass

    # Sobreescribimos 'solution()'
    def solution(self) -> Solution:
        return self.score(), self.decisions()


# Solver de minimización (p.e. para el problema del cambio)
def bt_min_solve(initial_ds: ScoredDecisionSequence[TDecision]) -> Iterator[Solution]:
    def bt(ds: ScoredDecisionSequence[TDecision]) -> Iterator[Solution]:
        nonlocal bs
        ds_score = ds.score()
        best_seen[ds.state()] = ds_score
        if ds.is_solution() and ds_score < bs:  # sólo muestra una solución si mejora la última mostrada
            bs = ds_score
            yield ds.solution()
        for new_ds in ds.successors():
            new_state = new_ds.state()
            if new_ds.score() < best_seen.get(new_state, infinity):
                yield from bt(new_ds)

    best_seen = {}
    bs = infinity  # score of the best solution found
    return bt(initial_ds)


# Solver de maximización (p.e. para el problema de la mochila)
def bt_max_solve(initial_ds: ScoredDecisionSequence[TDecision]) -> Iterator[Solution]:
    def bt(ds: ScoredDecisionSequence[TDecision]) -> Iterator[Solution]:
        nonlocal bs
        ds_score = ds.score()
        best_seen[ds.state()] = ds_score
        if ds.is_solution() and ds_score > bs:  # sólo muestra una solución si mejora la última mostrada
            bs = ds_score
            yield ds.solution()
        for new_ds in ds.successors():
            new_state = new_ds.state()
            if new_ds.score() > best_seen.get(new_state, -infinity):
                yield from bt(new_ds)

    best_seen = {}
    bs = -infinity
    return bt(initial_ds)



from random import seed, randint

Capacity = int
Weight = int
Value = int
Score = int
LParams = tuple[int, Capacity]

Decision = int  # 0 o 1


def solve(c: Capacity,
          w: list[Weight],
          v: list[Value]) -> tuple[Score, list[Decision]]:
    def L(n: int, c: Capacity) -> Score:
        if n == 0:
            return 0
        if (n, c) not in mem:
            c_score: Score = L(n - 1, c)
            parent: LParams = n - 1, c
            decision: Decision = 0
            mem[n, c] = (c_score, parent, decision)
            if w[n - 1] <= c:
                c_score = L(n - 1, c - w[n - 1]) + v[n - 1]
                parent = n - 1, c - w[n - 1]
                decision = 1
                mem[n, c] = max(mem[n, c], (c_score, parent, decision))
        return mem[n, c][0]

    mem: dict[LParams, tuple[Score, LParams, Decision]] = {}
    score = L(len(v), c)
    n = len(v)
    sol = []
    while n > 0:
        _, (n, c), d = mem[n, c]
        sol.append(d)
    sol.reverse()
    return score, sol


Decision = int  # Indice de objeto


def solve2(c: Capacity,
           w: list[Weight],
           v: list[Value]) -> tuple[Score, list[Decision]]:
    def L(n: int, c: Capacity) -> Score:
        if n == 0:
            return 0
        if (n, c) not in mem:
            res = []
            for d in range(n):
                if w[d] <= c:
                    res.append((L(d, c - w[d]) + v[d], (d, c - w[d]), d))
            if len(res) == 0:
                mem[n, c] = 0, (0, c), -1  # -1 cuando no quepa ningun objeto
            else:
                mem[n, c] = max(res)
        return mem[n, c][0]

    mem: dict[LParams, tuple[Score, LParams, Decision]] = {}
    score = L(len(v), c)
    n = len(v)
    sol = []
    while n > 0:
        _, (n, c), d = mem[n, c]
        if d == -1: break  # No cabe ningún objeto
        sol.append(d)
    sol.reverse()
    return score, sol


def sorted_by_dec_ratio(w_old, v_old):
    idxs = sorted(range(len(w_old)), key=lambda i: -v_old[i] / w_old[i])
    w_new = [w_old[i] for i in idxs]
    v_new = [v_old[i] for i in idxs]
    return w_new, v_new


def create_knapsack_problem(num_objects):
    seed(5)
    w_new = [randint(10, 100) for _ in range(num_objects)]
    v_new = [w_new[i] * randint(1, 4) for i in range(num_objects)]
    capacity = int(sum(w_new) * 0.3)
    weights, values = sorted_by_dec_ratio(w_new, v_new)
    return weights, values, capacity


if __name__ == '__main__':
    w, v, c = [10, 5, 6, 2, 6], [10, 2, 3, 4, 2], 1  # Solution: value = 8,weight = 9,decisions = (1, 0, 1, 1, 0))
    # w, v, c = create_knapsack_problem(20)  # Solution: value = 1118, weight = 344, decisions = (1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0)
    # w, v, c = create_knapsack_problem(35)  # Solution: value = 1830, weight = 543, decisions = (1, 1, 1, 1, 1, 0, 1, 0, 1, 1, 1, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0)

    print("Instance:", c, w, v)
    print("Solution1:", solve(c, w, v))
    print("Solution2:", solve2(c, w, v))
